"""graph commands for cmem command line interface."""

import gzip
import hashlib
import io
import json
import mimetypes
import os
from xml.dom import minidom  # nosec
from xml.etree.ElementTree import (  # nosec
    Element,
    SubElement,
    tostring,
)

import click
from click import ClickException, Context, UsageError
from cmem.cmempy.config import get_cmem_base_uri
from cmem.cmempy.dp.authorization import refresh
from cmem.cmempy.dp.proxy import graph as graph_api
from cmem.cmempy.dp.proxy.graph import get_graph_imports
from cmem.cmempy.dp.proxy.sparql import get as sparql_api
from jinja2 import Template
from six.moves.urllib.parse import quote

from cmem_cmemc import completion
from cmem_cmemc.command import CmemcCommand
from cmem_cmemc.command_group import CmemcGroup
from cmem_cmemc.commands.graph_imports import graph_imports_list, imports_group
from cmem_cmemc.commands.validation import validation_group
from cmem_cmemc.constants import UNKNOWN_GRAPH_ERROR
from cmem_cmemc.context import ApplicationContext
from cmem_cmemc.parameter_types.path import ClickSmartPath
from cmem_cmemc.smart_path import SmartPath
from cmem_cmemc.utils import (
    RdfGraphData,
    convert_uri_to_filename,
    get_graphs,
    get_graphs_as_dict,
    iri_to_qname,
    read_rdf_graph_files,
    tuple_to_list,
)


def count_graph(graph_iri: str) -> int:
    """Count triples in a graph and return integer."""
    query = "SELECT (COUNT(*) AS ?triples) " + " FROM <" + graph_iri + "> WHERE { ?s ?p ?o }"  # noqa: S608
    result = json.loads(sparql_api(query, owl_imports_resolution=False))
    count = result["results"]["bindings"][0]["triples"]["value"]
    return int(count)


def _get_graph_to_file(  # noqa: PLR0913
    graph_iri: str,
    file_path: str,
    app: ApplicationContext,
    numbers: tuple[int, int] | None = None,
    overwrite: bool = True,
    mime_type: str = "application/n-triples",
) -> None:
    """Request a single graph to a single file (streamed).

    numbers is a tuple of current and count (for output only).
    """
    if SmartPath(file_path).exists():
        if overwrite is True:
            app.echo_warning(f"Output file {file_path} does exist: will overwrite it.")
        else:
            app.echo_warning(f"Output file {file_path} does exist: will append to it.")
    if numbers is not None:
        running_number, count = numbers
        if running_number is not None and count is not None:
            app.echo_info(
                f"Export graph {running_number}/{count}: " f"{graph_iri} to {file_path} ... ",
                nl=False,
            )
    # create and write the .ttl content file
    mode = "wb" if overwrite is True else "ab"

    with (
        gzip.open(file_path, mode)
        if file_path.endswith(".gz")
        else click.open_file(file_path, mode) as triple_file,
        graph_api.get_streamed(graph_iri, accept=mime_type) as response,
    ):
        response.raise_for_status()
        for chunk in response.iter_content(chunk_size=None):
            if chunk:
                triple_file.write(chunk)
        request_headers = response.request.headers
        request_headers.pop("Authorization")
        app.echo_debug(f"cmemc request headers: {request_headers}")
        app.echo_debug(f"server response headers: {response.headers}")
    if numbers is not None:
        app.echo_success("done")


def _get_export_names(
    app: ApplicationContext, iris: list[str], template: str, file_extension: str = ".ttl"
) -> dict:
    """Get a dictionary of generated file names based on a template.

    Args:
    ----
        app: the context click application
        iris: list of graph iris
        template (str): the template string to use
        file_extension(str): the file extension to use

    Returns:
    -------
        a dictionary with IRIs as keys and filenames as values

    Raises:
    ------
        ClickException in case the template string produces a naming clash,
            means two IRIs result in the same filename

    """
    template_data = app.get_template_data()
    _names = {}
    for iri in iris:
        template_data.update(
            hash=hashlib.sha256(iri.encode("utf-8")).hexdigest(),
            iriname=convert_uri_to_filename(iri),
        )
        _name_created = f"{Template(template).render(template_data)}{file_extension}"
        _names[iri] = _name_created
    if len(_names.values()) != len(set(_names.values())):
        raise ClickException(
            "The given template string produces a naming clash. "
            "Please use a different template to produce unique names."
        )
    return _names


def _get_graphs_filtered(filter_name: str, filter_value: str) -> list[dict]:
    """Get graphs but filtered according to name and value."""
    # not filtered means all graphs
    graphs: list[dict]
    if filter_name is None:
        return get_graphs()
    # check for correct filter names
    possible_filter_names = ("access", "imported-by")
    if filter_name not in possible_filter_names:
        raise ClickException(
            f"{filter_name} is an unknown filter name. " f"Use one of {possible_filter_names}."
        )
    # filter by access condition
    if filter_name == "access":
        if filter_value == "writeable":
            graphs = get_graphs(writeable=True, readonly=False)
        elif filter_value == "readonly":
            graphs = get_graphs(writeable=False, readonly=True)
        else:
            raise ClickException("Filter access is either 'readonly' or 'writeable'.")
    else:
        # default is all graphs
        graphs = get_graphs()
    # filter by imported-by
    if filter_name == "imported-by":
        if filter_value not in get_graphs_as_dict():
            raise ClickException(UNKNOWN_GRAPH_ERROR.format(filter_value))
        imported_graphs = get_graph_imports(filter_value)
        graphs = [_ for _ in graphs if _["iri"] in imported_graphs]
    return graphs


def _add_imported_graphs(iris: list[str], all_graphs: dict) -> list[str]:
    """Get a list of graph IRIs extended with the imported graphs.

    Args:
    ----
        iris: list of graph IRIs
        all_graphs: output of get_graphs_as_dict (dict of all graphs)

    Returns:
    -------
        list of graph IRIs

    """
    extended_list = iris
    for iri in set(iris):
        for _ in get_graph_imports(iri):
            # check if graph exist
            if _ in all_graphs:
                extended_list.append(_)
    return list(set(extended_list))


def _check_and_extend_exported_graphs(
    iris: list[str], all_flag: bool, imported_flag: bool, all_graphs: dict
) -> list[str]:
    """Get a list of IRIs checked and extended.

    Args:
    ----
        iris: List or tuple of given user IRIs
        all_flag: user wants all graphs
        imported_flag: user wants imported graphs as well
        all_graphs: dict of all graphs (get_graph_as_dict())

    Returns:
    -------
        checked and extended list of IRIs

    Raises:
    ------
         UsageError or ClickException

    """
    # transform given IRI-tuple to a distinct IRI-list
    iris = list(set(iris))
    if len(iris) == 0 and not all_flag:
        raise UsageError(
            "Either provide at least one graph IRI or use the --all option "
            "in order to work with all graphs."
        )
    for iri in iris:
        if iri not in all_graphs:
            raise ClickException(UNKNOWN_GRAPH_ERROR.format(iri))
    if all_flag:
        # in case --all is given,
        # list of graphs is filled with all available graph IRIs
        iris = [str(_) for _ in all_graphs]
    elif imported_flag:
        # does not need be executed in case of --all
        iris = _add_imported_graphs(iris, all_graphs)
    return iris


def _create_xml_catalog_file(app: ApplicationContext, names: dict, output_dir: str) -> None:
    """Create a Protégé suitable XML catalog file.

    Args:
    ----
        app: the cmemc context object
        names: output of _get_export_names()
        output_dir: path where to create the XML file

    """
    file_name = SmartPath(output_dir) / "catalog-v001.xml"
    catalog = Element("catalog")
    catalog.set("prefer", "public")
    catalog.set("xmlns", "urn:oasis:names:tc:entity:xmlns:xml:catalog")
    for name in names:
        uri = SubElement(catalog, "uri")
        uri.set("id", "Auto-Generated Import Resolution by cmemc")
        uri.set("name", name)
        uri.set("uri", names[name])
    parsed_string = minidom.parseString(  # nosec - since source is trusted  # noqa: S318
        tostring(catalog, "utf-8")
    ).toprettyxml(indent="  ")
    file = click.open_file(str(file_name), "w")
    file.truncate(0)
    file.write(parsed_string)
    app.echo_success(f"XML catalog file created as {file_name}.")


@click.command(cls=CmemcCommand, name="tree", hidden=True)
@click.option("-a", "--all", "all_", is_flag=True, help="Show tree of all (readable) graphs.")
@click.option("--raw", is_flag=True, help="Outputs raw JSON of the graph importTree API response.")
@click.option(
    "--id-only",
    is_flag=True,
    help="Lists only graph identifier (IRIs) and no labels or other "
    "metadata. This is useful for piping the IRIs into other commands. "
    "The output with this option is a sorted, flat, de-duplicated list "
    "of existing graphs.",
)
@click.argument(
    "iris",
    nargs=-1,
    type=click.STRING,
    shell_complete=completion.graph_uris,
    callback=tuple_to_list,
)
@click.pass_context
def tree_command(ctx: Context, all_: bool, raw: bool, id_only: bool, iris: list[str]) -> None:
    """(Hidden) Deprecated: use 'graph imports tree' instead."""
    app: ApplicationContext = ctx.obj
    app.echo_warning(
        "The 'tree' command is deprecated and will be removed with the next major release."
        " Please use the 'graph imports tree' command instead."
    )
    imports_cmd = graph.commands["imports"]
    tree_cmd = imports_cmd.commands["tree"]  # type: ignore[attr-defined]
    ctx.invoke(tree_cmd, all_=all_, raw=raw, id_only=id_only, iris=iris)


@click.command(cls=CmemcCommand, name="list")
@click.option("--raw", is_flag=True, help="Outputs raw JSON of the graphs list API response.")
@click.option(
    "--id-only",
    is_flag=True,
    help="Lists only graph identifier (IRIs) and no labels or other "
    "metadata. This is useful for piping the IRIs into other commands.",
)
@click.option(
    "--filter",
    "filter_",
    type=click.Tuple([click.Choice(["access", "imported-by"]), str]),
    shell_complete=completion.graph_list_filter,
    default=[None] * 2,
    help="Filter graphs based on effective access conditions or import "
    "closure. "
    "First parameter CHOICE can be 'access' or 'imported-by'. "
    "The second parameter can be 'readonly' or 'writeable' in case "
    "of 'access' or any readable graph in case of 'imported-by'.",
)
@click.pass_obj
def list_command(
    app: ApplicationContext, raw: bool, id_only: bool, filter_: tuple[str, str]
) -> None:
    """List accessible graphs."""
    filter_name, filter_value = filter_
    graphs = _get_graphs_filtered(filter_name, filter_value)

    if raw:
        app.echo_info_json(graphs)
        return
    if id_only:
        # output a sorted list of graph IRIs
        for graph_desc in sorted(graphs, key=lambda k: k["iri"].lower()):
            app.echo_result(graph_desc["iri"])
        return
    # output a user table
    table = []
    for _ in graphs:
        if len(_["assignedClasses"]) > 0:
            graph_class = iri_to_qname(sorted(_["assignedClasses"])[0])
        else:
            graph_class = ""
        row = [
            _["iri"],
            graph_class,
            _["label"]["title"],
        ]
        table.append(row)
    app.echo_info_table(
        table,
        headers=["Graph IRI", "Type", "Label"],
        sort_column=2,
        empty_table_message="No graphs found. "
        "Use the `graph import` command to import a graph from a file, or "
        "use the `admin store bootstrap` command to import the default graphs.",
    )


def _validate_export_command_input_parameters(
    output_dir: str, output_file: str, compress: str, create_catalog: bool
) -> None:
    """Validate export command input parameters combinations"""
    if output_dir and create_catalog and compress:
        raise UsageError(
            "Cannot create a catalog file when using a compressed graph file."
            " Please remove either the --create-catalog or --compress option."
        )
    if output_file == "- " and compress:
        raise UsageError("Cannot output a binary file to terminal. Use --output-file option.")


def _write_graph_imports(ctx: Context, filename: str, iri: str) -> None:
    imports = graph_imports_list.apply_filters(ctx=ctx, filter_=[("to-graph", iri)])
    if imports:
        imports_file = click.open_file(filename, "w")
        for _ in imports:
            imports_file.write(_["from_graph"] + "\n")
        imports_file.close()


# pylint: disable=too-many-arguments,too-many-locals
@click.command(cls=CmemcCommand, name="export")
@click.option("-a", "--all", "all_", is_flag=True, help="Export all readable graphs.")
@click.option(
    "--include-imports",
    is_flag=True,
    help="Export selected graph(s) and all graphs which are imported from "
    "these selected graph(s).",
)
@click.option(
    "--include-import-statements",
    is_flag=True,
    help="Save graph imports information from other graphs to the exported graphs "
    "and write *.imports files.",
)
@click.option(
    "--create-catalog",
    is_flag=True,
    help="In addition to the .ttl and .graph files, cmemc will create an "
    "XML catalog file (catalog-v001.xml) which can be used by "
    "applications such as Protégé.",
)
@click.option(
    "--output-dir",
    type=ClickSmartPath(writable=True, file_okay=False),
    help="Export to this directory.",
)
@click.option(
    "--output-file",
    type=ClickSmartPath(writable=True, allow_dash=True, dir_okay=False),
    default="-",
    show_default=True,
    shell_complete=completion.triple_files,
    help="Export to this file.",
)
@click.option(
    "--filename-template",
    "-t",
    "template",
    default="{{hash}}",
    show_default=True,
    type=click.STRING,
    shell_complete=completion.graph_export_templates,
    help="Template for the export file name(s). "
    "Used together with --output-dir. "
    "Possible placeholders are (Jinja2): "
    "{{hash}} - sha256 hash of the graph IRI, "
    "{{iriname}} - graph IRI converted to filename, "
    "{{connection}} - from the --connection option and "
    "{{date}} - the current date as YYYY-MM-DD. "
    "The file suffix will be appended. "
    "Needed directories will be created.",
)
@click.option(
    "--mime-type",
    default="text/turtle",
    show_default=True,
    type=click.Choice(["application/n-triples", "text/turtle", "application/rdf+xml"]),
    help="Define the requested mime type",
)
@click.option(
    "--compress",
    type=click.Choice(["gzip"]),
    help="Compress the exported graph files.",
)
@click.argument(
    "iris",
    nargs=-1,
    type=click.STRING,
    shell_complete=completion.graph_uris,
    callback=tuple_to_list,
)
@click.pass_context
def export_command(  # noqa: C901, PLR0913
    ctx: Context,
    all_: bool,
    include_imports: bool,
    include_import_statements: bool,
    create_catalog: bool,
    output_dir: str,
    output_file: str,
    template: str,
    mime_type: str,
    iris: list[str],
    compress: str,
) -> None:
    """Export graph(s) as NTriples to stdout (-), file or directory.

    In case of file export, data from all selected graphs will be concatenated
    in one file.
    In case of directory export, .graph and .ttl files will be created
    for each graph.
    """
    app: ApplicationContext = ctx.obj
    _validate_export_command_input_parameters(output_dir, output_file, compress, create_catalog)
    iris = _check_and_extend_exported_graphs(iris, all_, include_imports, get_graphs_as_dict())

    count: int = len(iris)
    app.echo_debug("graph count is " + str(count))
    if output_dir:
        # output directory set
        app.echo_debug("output is directory")
        # pre-calculate all filenames with the template,
        # in order to output errors on naming clashes as early as possible
        extension = mimetypes.guess_extension(mime_type)
        _names = _get_export_names(
            app, iris, template, f"{extension}.gz" if compress else f"{extension}"
        )
        _graph_file_names = _get_export_names(app, iris, template, f"{extension}.graph")
        _graph_imports_file_names = _get_export_names(app, iris, template, f"{extension}.imports")

        # create directory
        if not SmartPath(output_dir).exists():
            app.echo_warning("Output directory does not exist: " + "will create it.")
            SmartPath(output_dir).mkdir(parents=True)
        # one .graph, one .ttl file per named graph
        for current, iri in enumerate(iris, start=1):
            # join with given output directory and normalize full path
            triple_file_name = os.path.normpath(SmartPath(output_dir) / _names[iri])
            graph_file_name = os.path.normpath(SmartPath(output_dir) / _graph_file_names[iri])
            imports_file_name = os.path.normpath(
                SmartPath(output_dir) / _graph_imports_file_names[iri]
            )
            # output directory is created lazy
            SmartPath(triple_file_name).parent.mkdir(parents=True, exist_ok=True)
            # create and write the .ttl.graph metadata file
            graph_file = click.open_file(graph_file_name, "w")
            graph_file.write(iri + "\n")
            graph_file.close()
            if include_import_statements:
                _write_graph_imports(ctx=ctx, filename=imports_file_name, iri=iri)
            _get_graph_to_file(
                iri, triple_file_name, app, numbers=(current, count), mime_type=mime_type
            )

        if create_catalog:
            _create_xml_catalog_file(app, _names, output_dir)
        return
    # no output directory set -> file export
    if output_file == "-":
        if compress:
            raise UsageError("Cannot output a binary file to terminal. Use --output-file option.")
        # in case a file is stdout,
        # all triples from all graphs go in and other output is suppressed
        app.echo_debug("output is stdout")
        for iri in iris:
            _get_graph_to_file(iri, output_file, app, mime_type=mime_type)
    else:
        # in case a file is given, all triples from all graphs go in
        if compress and not output_file.endswith(".gz"):
            output_file = output_file + ".gz"

        app.echo_debug("output is file")
        for current, iri in enumerate(iris, start=1):
            _get_graph_to_file(
                iri,
                output_file,
                app,
                numbers=(current, count),
                overwrite=False,
                mime_type=mime_type,
            )


def validate_input_path(input_path: str) -> None:
    """Validate input path

    This function checks the provided folder for any .ttl or .nt files
    that have corresponding .gz files. If such files are found, it raises a ClickException.
    """
    files = os.listdir(input_path)

    # Check for files with the given extensions (.ttl and .nt)
    rdf_files = [f for f in files if f.endswith((".ttl", ".nt"))]

    # Check for corresponding .gz files
    gz_files = [f"{f}.gz" for f in rdf_files]
    conflicting_files = [f for f in gz_files if f in files]

    if conflicting_files:
        raise ClickException(
            f"The following RDF files (.ttl/.nt) have corresponding '.gz' files,"
            f" which is not allowed: {', '.join(conflicting_files)}"
        )


def _get_graph_supported_formats() -> dict[str, str]:
    return {
        "application/rdf+xml": "xml",
        "application/ld+json": "jsonld",
        "text/turtle": "turtle",
        "application/n-triples": "nt",
    }


def _get_buffer_and_content_type(
    triple_file: str, app: ApplicationContext
) -> tuple[io.BytesIO, str]:
    """Get the io.BytesIO buffer and the content type of triple_file"""
    smart_file = SmartPath(triple_file)
    content_type, encoding = mimetypes.guess_type(triple_file)
    if content_type is None:
        content_type = "text/turtle"
        for supported_type, supported_suffix in _get_graph_supported_formats().items():
            if smart_file.name.endswith(f".{supported_suffix}") or smart_file.name.endswith(
                f".{supported_suffix}.gz"
            ):
                content_type = supported_type
    elif content_type not in _get_graph_supported_formats():
        app.echo_warning(
            f"Content type {content_type} of {triple_file} is "
            f"not one of {', '.join(_get_graph_supported_formats().keys())} "
            "(but will try to import anyways)."
        )

    transport_params = {}
    if smart_file.schema in ["http", "https"]:
        transport_params["headers"] = {
            "Accept": "text/turtle; q=1.0, application/x-turtle; q=0.9, text/n3;"
            " q=0.8, application/rdf+xml; q=0.5, text/plain; q=0.1"
        }

    buffer = io.BytesIO()
    with ClickSmartPath.open(triple_file, transport_params=transport_params) as file_obj:
        buffer.write(file_obj.read())
    buffer.seek(0)
    return buffer, content_type


def _create_graph_imports(ctx: Context, graphs: list[RdfGraphData]) -> None:
    # Locate and invoke the 'create' subcommand under 'graph imports' command
    app: ApplicationContext = ctx.obj
    imports_cmd = graph.commands["imports"]
    create_cmd = imports_cmd.commands["create"]  # type: ignore[attr-defined]
    for _ in graphs:
        if not _.graph_imports:
            continue
        for _import in _.graph_imports:
            from_graph = _import
            to_graph = _.graph_iri
            imports = graph_imports_list.apply_filters(
                ctx=ctx, filter_=[("from-graph", from_graph), ("to-graph", to_graph)]
            )
            if imports:
                app.echo_info(
                    f"Creating graph import from {from_graph} to {to_graph} ... ", nl=False
                )
                app.echo_warning("exists")
            else:
                ctx.invoke(create_cmd, from_graph=from_graph, to_graph=to_graph)


def _process_input_directory(input_path: str, iri: str) -> list[RdfGraphData]:
    if iri is None:
        # in case a directory is the source (and no IRI is given),
        # the graph/nt file structure is crawled
        graphs = read_rdf_graph_files(input_path)
    else:
        # in case a directory is the source AND IRI is given
        graphs = []
        for _ in _get_graph_supported_formats():
            extension = mimetypes.guess_extension(_)
            graphs += [
                RdfGraphData(str(file), iri, [])
                for file in SmartPath(input_path).glob(f"*{extension}")
            ]
            graphs += [
                RdfGraphData(str(file), iri, [])
                for file in SmartPath(input_path).glob(f"*{extension}.gz")
            ]
    return graphs


def _validate_graph_imports(graphs_to_import: list[RdfGraphData]) -> None:
    graphs = {_["iri"] for _ in get_graphs()}
    graphs.update({_.graph_iri for _ in graphs_to_import})
    graph_imports: set[str] = set()
    for graph_data in graphs_to_import:
        for item in graph_data.graph_imports:
            graph_imports.add(str(item))
    if graph_imports - graphs:
        raise click.UsageError(
            f"The following graphs are not available, "
            f"so you can not add import statements to them: {','.join(graph_imports - graphs)}"
        )


@click.command(cls=CmemcCommand, name="import")
@click.option(
    "--replace",
    is_flag=True,
    help="Replace / overwrite the graph(s), instead of just adding the triples to the graph.",
)
@click.option(
    "--skip-existing",
    is_flag=True,
    help="Skip importing a file if the target graph already exists in "
    "the store. Note that the graph list is fetched once at the "
    "beginning of the process, so that you can still add multiple "
    "files to one single graph (if it does not exist).",
)
@click.option(
    "--include-import-statements",
    is_flag=True,
    help="Use *.imports files to re-apply the graph imports of the imported graphs.",
)
@click.argument(
    "input_path",
    required=True,
    shell_complete=completion.triple_files,
    type=ClickSmartPath(allow_dash=False, readable=True, remote_okay=True),
)
@click.argument("iri", type=click.STRING, required=False, shell_complete=completion.graph_uris)
@click.pass_context
def import_command(  # noqa: PLR0913
    ctx: Context,
    input_path: str,
    replace: bool,
    skip_existing: bool,
    include_import_statements: bool,
    iri: str,
) -> None:
    """Import graph(s) to the store.

    If input is a file, content will be uploaded to the graph identified with the IRI.

    If input is a directory and NO IRI is given, it scans for file-pairs such as
    `xyz.ttl` and `xyz.ttl.graph`, where `xyz.ttl` is the actual triples file and
    `xyz.ttl.graph` contains the graph IRI in the first line: `https://mygraph.de/xyz/`.

    If input is a directory AND a graph IRI is given, it scans for all `*.ttl` files
    in the directory and imports all content to the graph, ignoring the `*.ttl.graph`
    files.

    If the `--replace` flag is set, the data in the graphs will be overwritten,
    if not, it will be added.

    Note: Directories are scanned on the first level only (not recursively).
    """
    app: ApplicationContext = ctx.obj
    if replace and skip_existing:
        raise UsageError(
            "The options --replace and --skip-existing are mutually "
            "exclusive, so please remove one of them."
        )
    # is an array of tuples like this [('path/to/triple.file', 'graph IRI')]
    graphs: list[RdfGraphData]
    if SmartPath(input_path).is_dir():
        validate_input_path(input_path)
        graphs = _process_input_directory(input_path, iri)
        if include_import_statements:
            _validate_graph_imports(graphs)
    elif SmartPath(input_path).is_file():
        if iri is None:
            raise UsageError(
                "Either specify an input file AND a graph IRI or an input directory ONLY."
            )
        graphs = [RdfGraphData(input_path, iri, [])]
    else:
        raise NotImplementedError(
            "Input from special files (socket, FIFO, device file) is not supported."
        )

    existing_graphs = get_graphs_as_dict()
    processed_graphs: set = set()
    count: int = len(graphs)
    current: int = 1
    for _ in graphs:
        triple_file = _.file_path
        graph_iri = _.graph_iri
        app.echo_info(
            f"Import file {current}/{count}: " f"{graph_iri} from {triple_file} ... ", nl=False
        )
        if graph_iri in existing_graphs and skip_existing:
            app.echo_warning("skipped")
            continue
        # prevents re-replacing of graphs in a single run
        _replace = False if graph_iri in processed_graphs else replace
        _buffer, content_type = _get_buffer_and_content_type(triple_file, app)
        response = graph_api.post_streamed(
            graph_iri, _buffer, replace=_replace, content_type=content_type
        )
        request_headers = response.request.headers
        request_headers.pop("Authorization")
        app.echo_debug(f"cmemc request headers: {request_headers}")
        app.echo_debug(f"server response headers: {response.headers}")
        app.echo_success("replaced" if _replace else "added")
        # refresh access conditions in case of dropped AC graph
        if graph_iri == refresh.AUTHORIZATION_GRAPH_URI:
            refresh.get()
            app.echo_debug("Access conditions refreshed.")
        processed_graphs.add(graph_iri)
        current += 1
    # create graph imports
    if include_import_statements:
        _create_graph_imports(ctx, graphs)


@click.command(cls=CmemcCommand, name="delete")
@click.option("-a", "--all", "all_", is_flag=True, help="Delete all writeable graphs.")
@click.option(
    "--include-imports",
    is_flag=True,
    help="Delete selected graph(s) and all writeable graphs which are "
    "imported from these selected graph(s).",
)
@click.option(
    "--include-import-statements", is_flag=True, help="Delete import reference of deleted graphs"
)
@click.argument(
    "iris",
    nargs=-1,
    type=click.STRING,
    shell_complete=completion.writable_graph_uris,
    callback=tuple_to_list,
)
@click.pass_context
def delete_command(
    ctx: Context,
    all_: bool,
    include_imports: bool,
    include_import_statements: bool,
    iris: list[str],
) -> None:
    """Delete graph(s) from the store."""
    app: ApplicationContext = ctx.obj
    graphs = get_graphs_as_dict(writeable=True, readonly=False)
    iris = _check_and_extend_exported_graphs(iris, all_, include_imports, graphs)
    imports_to_be_deleted = []
    count: int = len(iris)
    for current, iri in enumerate(iris, start=1):
        imports_to_be_deleted += graph_imports_list.apply_filters(
            ctx=ctx, filter_=[("to-graph", iri)]
        )

        app.echo_info(f"Drop graph {current}/{count}: {iri} ... ", nl=False)
        graph_api.delete(iri)
        app.echo_success("done")
        # refresh access conditions in case of dropped AC graph
        if iri == refresh.AUTHORIZATION_GRAPH_URI:
            refresh.get()
            app.echo_debug("Access conditions refreshed.")
    if include_import_statements:
        # Locate and invoke the 'delete' subcommand under 'graph imports' command
        imports_cmd = graph.commands["imports"]
        delete_cmd = imports_cmd.commands["delete"]  # type: ignore[attr-defined]
        for _ in imports_to_be_deleted:
            if _["from_graph"] not in iris:
                ctx.invoke(delete_cmd, from_graph=_["from_graph"], to_graph=_["to_graph"])


@click.command(cls=CmemcCommand, name="open")
@click.argument("iri", type=click.STRING, shell_complete=completion.graph_uris)
@click.pass_obj
def open_command(app: ApplicationContext, iri: str) -> None:
    """Open / explore a graph in the browser."""
    explore_uri = get_cmem_base_uri() + "/explore?graph=" + quote(iri)
    click.launch(explore_uri)
    app.echo_debug(explore_uri)


@click.command(cls=CmemcCommand, name="count")
@click.option("-a", "--all", "all_", is_flag=True, help="Count all graphs")
@click.option(
    "-s", "--summarize", is_flag=True, help="Display only a sum of all counted graphs together"
)
@click.argument("iris", nargs=-1, type=click.STRING, shell_complete=completion.graph_uris)
@click.pass_obj
def count_command(
    app: ApplicationContext, all_: bool, summarize: bool, iris: tuple[str, ...]
) -> None:
    """Count triples in graph(s).

    This command lists graphs with their triple count.
    Counts do not include imported graphs.
    """
    if not iris and not all_:
        raise UsageError(
            "Either specify at least one graph IRI " "or use the --all option to count all graphs."
        )
    if all_:
        # in case --all is given,
        # list of graphs is filled with all available graph IRIs
        iris = tuple(iri["iri"] for iri in get_graphs())

    count: int
    overall_sum: int = 0
    for iri in iris:
        count = count_graph(iri)
        overall_sum = overall_sum + count
        if not summarize:
            app.echo_result(f"{count!s} {iri}")
    if summarize:
        app.echo_result(str(overall_sum))


@click.group(cls=CmemcGroup)
def graph() -> CmemcGroup:  # type: ignore[empty-body]
    """List, import, export, delete, count, tree or open graphs.

    Graphs are identified by an IRI.

    Note: The get a list of existing graphs,
    execute the `graph list` command or use tab-completion.
    """


graph.add_command(count_command)
graph.add_command(tree_command)
graph.add_command(list_command)
graph.add_command(export_command)
graph.add_command(delete_command)
graph.add_command(import_command)
graph.add_command(open_command)
graph.add_command(validation_group)
graph.add_command(imports_group)



"""Helper functions for rich text output"""

from abc import ABC, abstractmethod
from datetime import datetime, timezone
from urllib.parse import quote

import timeago
from cmem.cmempy.config import get_cmem_base_uri

from cmem_cmemc.title_helper import TitleHelper
from cmem_cmemc.utils import get_graphs_as_dict


class StringProcessor(ABC):
    """ABC of a table cell string processor"""

    @abstractmethod
    def process(self, text: str) -> str:
        """Process a single string content and output the processed string."""


class TimeAgo(StringProcessor):
    """Create a string similar to 'x minutes ago' from a timestamp"""

    def process(self, text: str) -> str:
        """Process a single string content and output the processed string."""
        if text is None:
            return ""
        try:
            text_as_int = int(text)
        except ValueError:
            return text
        stamp = datetime.fromtimestamp(text_as_int / 1000, tz=timezone.utc)
        return str(timeago.format(stamp, datetime.now(tz=timezone.utc)))


class GraphLink(StringProcessor):
    """Create a graph link from an IRI cell

    "Visit my [link=https://www.willmcgugan.com]blog[/link]!"
    """

    def __init__(self) -> None:
        self.cmem_base_uri = get_cmem_base_uri()
        self.base = self.cmem_base_uri + "/explore?graph="
        self.graph_labels: dict[str, str] = {}
        for _ in get_graphs_as_dict().values():
            self.graph_labels[_["iri"]] = _["label"]["title"]

    def process(self, text: str) -> str:
        """Process a single string content and output the processed string."""
        link = self.base + quote(text)
        label = self.graph_labels.get(text, None)
        return f"[link={link}]{label}[/link]" if label else text


class ResourceLink(StringProcessor):
    """Create a resource link from an IRI cell

    "Visit my [link=https://www.willmcgugan.com]blog[/link]!"
    """

    def __init__(self, graph_iri: str, title_helper: TitleHelper | None = None):
        self.graph_iri = graph_iri
        self.base = get_cmem_base_uri() + "/explore?graph=" + quote(graph_iri) + "&resource="
        self.title_helper = title_helper if title_helper else TitleHelper()

    def process(self, text: str) -> str:
        """Process a single string content and output the processed string."""
        link = self.base + quote(text)
        label = self.title_helper.get(text)
        return f"[link={link}]{label}[/link]"


def process_row(row: list[str], hints: dict[int, StringProcessor]) -> list[str]:
    """Process all cells in a row according to the StringProcessors"""
    processed_row = []
    for column_number, cell in enumerate(row):
        if hints.get(column_number):
            processed_row.append(hints[column_number].process(cell))
        else:
            processed_row.append(cell)
    return processed_row



"""Build dataset resource commands for cmemc."""

import re

import click
from click import ClickException, UsageError
from cmem.cmempy.workspace.projects.resources import get_all_resources
from cmem.cmempy.workspace.projects.resources.resource import (
    delete_resource,
    get_resource_metadata,
    get_resource_usage_data,
)

from cmem_cmemc import completion
from cmem_cmemc.command import CmemcCommand
from cmem_cmemc.command_group import CmemcGroup
from cmem_cmemc.context import ApplicationContext
from cmem_cmemc.utils import split_task_id, struct_to_table

RESOURCE_FILTER_TYPES = ["project", "regex"]
RESOURCE_FILTER_TYPES_HIDDEN = ["ids"]
RESOURCE_FILTER_TEXT = (
    "Filter file resources based on metadata. "
    f"First parameter CHOICE can be one of {RESOURCE_FILTER_TYPES!s}"
    ". The second parameter is based on CHOICE, e.g. a project "
    "ID or a regular expression string."
)


def _get_resources_filtered(
    resources: list[dict], filter_name: str, filter_value: str | tuple[str, ...]
) -> list[dict]:
    """Get file resources but filtered according to name and value."""
    # check for correct filter names (filter ids is used internally only)
    if filter_name not in RESOURCE_FILTER_TYPES + RESOURCE_FILTER_TYPES_HIDDEN:
        raise UsageError(
            f"{filter_name} is an unknown filter name. " f"Use one of {RESOURCE_FILTER_TYPES}."
        )
    # filter by ID list
    if filter_name == "ids":
        return [_ for _ in resources if _["id"] in filter_value]
    # filter by project
    if filter_name == "project":
        return [_ for _ in resources if _["project"] == str(filter_value)]
    # filter by regex
    if filter_name == "regex":
        return [_ for _ in resources if re.search(str(filter_value), _["name"])]
    # return unfiltered list
    return resources


@click.command(cls=CmemcCommand, name="list")
@click.option("--raw", is_flag=True, help="Outputs raw JSON.")
@click.option(
    "--id-only",
    is_flag=True,
    help="Lists only resource names and no other metadata. "
    "This is useful for piping the IDs into other commands.",
)
@click.option(
    "--filter",
    "filters_",
    multiple=True,
    type=(str, str),
    shell_complete=completion.resource_list_filter,
    help=RESOURCE_FILTER_TEXT,
)
@click.pass_obj
def list_command(
    app: ApplicationContext, raw: bool, id_only: bool, filters_: tuple[tuple[str, str], ...]
) -> None:
    """List available file resources.

    Outputs a table or a list of dataset resources (files).
    """
    resources = get_all_resources()
    for _ in filters_:
        filter_name, filter_value = _
        resources = _get_resources_filtered(resources, filter_name, filter_value)
    if raw:
        app.echo_info_json(resources)
        return
    if id_only:
        for _ in sorted(_["id"] for _ in resources):
            app.echo_result(_)
        return
    # output a user table
    table = []
    headers = ["ID", "Modified", "Size"]
    for _ in resources:
        row = [
            _["id"],
            _["modified"],
            _["size"],
        ]
        table.append(row)
    app.echo_info_table(
        table,
        headers=headers,
        sort_column=0,
        empty_table_message="No dataset resources found. "
        "Use the `dataset create` command to create a new file based dataset.",
    )


@click.command(cls=CmemcCommand, name="delete")
@click.argument("resource_ids", nargs=-1, type=click.STRING, shell_complete=completion.resource_ids)
@click.option("--force", is_flag=True, help="Delete resource even if in use by a task.")
@click.option(
    "-a",
    "--all",
    "all_",
    is_flag=True,
    help="Delete all resources. " "This is a dangerous option, so use it with care.",
)
@click.option(
    "--filter",
    "filters_",
    multiple=True,
    type=(str, str),
    shell_complete=completion.resource_list_filter,
    help=RESOURCE_FILTER_TEXT,
)
@click.pass_obj
def delete_command(
    app: ApplicationContext,
    resource_ids: tuple[str, ...],
    force: bool,
    all_: bool,
    filters_: tuple[tuple[str, str], ...],
) -> None:
    """Delete file resources.

    There are three selection mechanisms: with specific IDs, only those
    specified resources will be deleted; by using --filter, resources based
    on the filter type and value will be deleted; using --all will delete
    all resources.
    """
    if resource_ids == () and not all_ and filters_ == ():
        raise UsageError(
            "Either specify at least one resource ID or use the --all or "
            "--filter options to specify resources for deletion."
        )

    resources = get_all_resources()
    if len(resource_ids) > 0:
        for resource_id in resource_ids:
            if resource_id not in [_["id"] for _ in resources]:
                raise ClickException(f"Resource {resource_id} not available.")
        # "filter" by id
        resources = _get_resources_filtered(resources, "ids", resource_ids)
    for _ in filters_:
        resources = _get_resources_filtered(resources, _[0], _[1])

    # avoid double removal as well as sort IDs
    processed_ids = sorted({_["id"] for _ in resources}, key=lambda v: v.lower())
    count = len(processed_ids)
    for current, resource_id in enumerate(processed_ids, start=1):
        current_string = str(current).zfill(len(str(count)))
        app.echo_info(f"Delete resource {current_string}/{count}: {resource_id} ... ", nl=False)
        project_id, resource_local_id = split_task_id(resource_id)
        usage = get_resource_usage_data(project_id, resource_local_id)
        if len(usage) > 0:
            app.echo_error(f"in use by {len(usage)} task(s)", nl=False)
            if force:
                app.echo_info(" ... ", nl=False)
            else:
                app.echo_info("")
                continue
        delete_resource(project_name=project_id, resource_name=resource_local_id)
        app.echo_success("deleted")


@click.command(cls=CmemcCommand, name="inspect")
@click.argument("resource_id", type=click.STRING, shell_complete=completion.resource_ids)
@click.option("--raw", is_flag=True, help="Outputs raw JSON.")
@click.pass_obj
def inspect_command(app: ApplicationContext, resource_id: str, raw: bool) -> None:
    """Display all metadata of a file resource."""
    project_id, resource_id = split_task_id(resource_id)
    resource_data = get_resource_metadata(project_id, resource_id)
    if raw:
        app.echo_info_json(resource_data)
    else:
        table = struct_to_table(resource_data)
        app.echo_info_table(table, headers=["Key", "Value"], sort_column=0)


@click.command(cls=CmemcCommand, name="usage")
@click.argument("resource_id", type=click.STRING, shell_complete=completion.resource_ids)
@click.option("--raw", is_flag=True, help="Outputs raw JSON.")
@click.pass_obj
def usage_command(app: ApplicationContext, resource_id: str, raw: bool) -> None:
    """Display all usage data of a file resource."""
    project_id, resource_id = split_task_id(resource_id)
    usage = get_resource_usage_data(project_id, resource_id)
    if raw:
        app.echo_info_json(usage)
        return
    # output a user table
    table = []
    headers = ["Task ID", "Type", "Label"]
    for _ in usage:
        row = [project_id + ":" + _["id"], _["taskType"], _["label"]]
        table.append(row)
    app.echo_info_table(table, headers=headers, sort_column=2)


@click.group(cls=CmemcGroup)
def resource() -> CmemcGroup:  # type: ignore[empty-body]
    """List, inspect or delete dataset file resources.

    File resources are identified by their paths and project IDs.
    """


resource.add_command(list_command)
resource.add_command(delete_command)
resource.add_command(inspect_command)
resource.add_command(usage_command)



"""Migration recipe abstract base class"""

import re
from abc import ABC, abstractmethod
from typing import ClassVar, Literal

from cmem.cmempy.health import get_complete_status_info
from cmem.cmempy.queries import SparqlQuery
from packaging.version import Version

components = Literal["explore", "build"]


class MigrationRecipe(ABC):
    """Tests and migration functions to migrate a specific type of resource between versions"""

    id: str
    description: str
    component: components = "explore"
    first_version: str | None = None
    last_version: str | None = None
    tags: ClassVar[list[str]] = []
    default_prefixes: ClassVar[dict[str, str]] = {
        "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
        "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
        "sh": "http://www.w3.org/ns/shacl#",
        "auth": "https://vocab.eccenca.com/auth/",
        "shui": "https://vocab.eccenca.com/shui/",
        "xsd": "http://www.w3.org/2001/XMLSchema#",
    }

    @abstractmethod
    def is_applicable(self) -> bool:
        """Test if the recipe can be applied."""

    @abstractmethod
    def apply(self) -> None:
        """Apply the recipe to the current version."""

    def version_matches(self, component_version: str = "") -> bool:
        """Test if the component version matches the version specification of the recipe"""
        status_info = get_complete_status_info()
        first_version = self.first_version if self.first_version else "0.0.0"
        last_version = self.last_version if self.last_version else "1000.0.0"
        if component_version == "":
            if self.component == "explore":
                component_version = str(status_info["explore"]["version"])
            if self.component == "build":
                component_version = str(status_info["build"]["version"])
        if component_version.startswith("v"):
            component_version = component_version[1:]
        # remove git describe stuff (24.2.1-501-gb0ca19b14)
        component_version = re.sub(r"-.*", "", component_version)
        if component_version == "":
            return False
        return Version(first_version) <= Version(component_version) <= Version(last_version)

    def _get_default_prefixes(self) -> str:
        """Get default prefixes as SPARQL Snippet"""
        snippet = ""
        for prefix, namespace in self.default_prefixes.items():
            snippet += f"PREFIX {prefix}: <{namespace}>\n"
        return snippet

    def _select(self, query_text: str, placeholder: dict[str, str] | None = None) -> list[dict]:
        """Get the bindings of a SPARQL Query result"""
        if not placeholder:
            placeholder = {"DEFAULT_PREFIXES": self._get_default_prefixes()}
        else:
            placeholder["DEFAULT_PREFIXES"] = self._get_default_prefixes()
        query = SparqlQuery(query_text)
        response: dict = query.get_json_results(placeholder=placeholder)
        results: dict = response.get("results", {"bindings": []})
        bindings: list[dict] = results.get("bindings")  # type: ignore[assignment]
        return bindings

    def _update(self, query_text: str, placeholder: dict[str, str] | None = None) -> None:
        """Send an update query"""
        if not placeholder:
            placeholder = {"DEFAULT_PREFIXES": self._get_default_prefixes()}
        else:
            placeholder["DEFAULT_PREFIXES"] = self._get_default_prefixes()
        query = SparqlQuery(query_text)
        query.query_type = "UPDATE"
        query.get_results(placeholder=placeholder)



"""Class to retrieve query placeholder resources

note: this is not the correct place for this class - it should be cmem_client or cmempy
"""

from cmem.cmempy.queries import SparqlQuery

FETCH_PLACEHOLDERS = """
PREFIX shui: <https://vocab.eccenca.com/shui/>
PREFIX rdf:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>

SELECT ?key ?valueQuery
FROM <{{graph}}>
WHERE {
  ?placeholder a shui:QueryPlaceholder ;
    shui:QueryPlaceholder_key ?key ;
    shui:QueryPlaceholder_valueQuery ?valueQueryR ;
    shui:QueryPlaceholder_usedInQuery <{{query}}> .
  ?valueQueryR shui:queryText ?valueQuery
}
"""


class QueryPlaceholder:
    """A query placeholder"""

    key: str
    value_query: SparqlQuery

    def __init__(self, key: str, value_query: str):
        self.key = key
        self.value_query = SparqlQuery(value_query, query_type="SELECT")

    def complete(self, incomplete: str = "") -> list[str] | list[tuple[str, str]]:  # noqa: ARG002
        """Prepare a list of placeholder values"""
        result = self.value_query.get_json_results()
        projection_vars = result["head"]["vars"]
        bindings = result["results"]["bindings"]
        if "value" not in projection_vars:
            return []

        if "description" not in projection_vars:
            values_without_description = []
            for _ in bindings:
                value = str(_["value"]["value"])
                values_without_description.append(value)
            return values_without_description

        values_with_description = []
        for _ in bindings:
            value = str(_["value"]["value"])
            description = str(_["description"]["value"])
            values_with_description.append((value, description))
        return values_with_description


def get_placeholders_for_query(iri: str) -> dict[str, QueryPlaceholder]:
    """Get a placeholder dict for a query"""
    placeholders = {}
    results = SparqlQuery(
        FETCH_PLACEHOLDERS,
        query_type="SELECT",
    ).get_json_results(placeholder={"graph": "https://ns.eccenca.com/data/queries/", "query": iri})
    for binding in results["results"]["bindings"]:
        key = binding["key"]["value"]
        value_query = binding["valueQuery"]["value"]
        placeholders[key] = QueryPlaceholder(key=key, value_query=value_query)
    return placeholders



"""cmemc Click Command"""

from subprocess import CalledProcessError
from typing import Any

import click
import requests
from click_help_colors import HelpColorsCommand

from cmem_cmemc.exceptions import CmemcError
from cmem_cmemc.utils import extract_error_message


class CmemcCommand(HelpColorsCommand):
    """Wrapper click.Command class to have a single extension point.

    Currently, wrapped click extensions and additional group features:#
    - click-help-colors: https://github.com/click-contrib/click-help-colors
    """

    color_for_headers = "yellow"
    color_for_options = "green"

    def __init__(self, *args, **kwargs):  # noqa: ANN002, ANN003
        """Init a cmemc group command."""
        kwargs.setdefault("help_headers_color", self.color_for_headers)
        kwargs.setdefault("help_options_color", self.color_for_options)
        super().__init__(*args, **kwargs)

    def invoke(self, ctx: click.core.Context) -> Any:  # noqa: ANN401
        """Execute the command and handles known exceptions by wrapping them in a CmemcError.

        This method overrides the default Click command invocation to catch a predefined
        set of exceptions (such as OSError, HTTPError, ValueError, etc.) and re-raises them
        as a unified CmemcError. This allows consistent error handling and messaging across
        all CLI commands.

        Args:
            ctx (click.core.Context): The Click context for the command.

        Raises:
            CmemcError: Wraps any of the caught exceptions and attaches context-specific data.

        """
        try:
            return super().invoke(ctx)
        except (
            OSError,
            CalledProcessError,
            requests.exceptions.HTTPError,
            requests.exceptions.ConnectionError,
            ValueError,
            NotImplementedError,
            KeyError,
        ) as e:
            raise CmemcError(ctx.obj, extract_error_message(e, True)) from e



"""Build (DataIntegration) python management commands."""

import sys
from dataclasses import asdict
from re import match

import click
from click import ClickException, UsageError
from cmem.cmempy.workspace.python import (
    install_package_by_file,
    install_package_by_name,
    list_packages,
    list_plugins,
    uninstall_package,
    update_plugins,
)

from cmem_cmemc import completion
from cmem_cmemc.command import CmemcCommand
from cmem_cmemc.command_group import CmemcGroup
from cmem_cmemc.context import ApplicationContext
from cmem_cmemc.parameter_types.path import ClickSmartPath
from cmem_cmemc.utils import get_published_packages


def _get_package_id(module_name: str) -> str:
    """Return package identifier."""
    return module_name.split(".")[0].replace("_", "-")


def _looks_like_a_package(package: str) -> bool:
    """Check if a string looks like a package requirement string."""
    return bool(match("^[a-zA-Z0-9]+(-[a-zA-Z0-9]+)*((==|<=|>=|>|<).*)?$", package))


@click.command(cls=CmemcCommand, name="install")
@click.argument(
    "PACKAGE",
    shell_complete=completion.installable_packages,
    type=ClickSmartPath(readable=True, allow_dash=False, dir_okay=False),
)
@click.pass_obj
def install_command(app: ApplicationContext, package: str) -> None:
    """Install a python package to the workspace.

    This command is essentially a `pip install` in the remote python
    environment.

    You can install a package by uploading a source distribution
    .tar.gz file, by uploading a build distribution .whl file, or by
    specifying a package name, i.e., a pip requirement specifier with a
    package name available on pypi.org (e.g. `requests==2.27.1`).

    Note: The tab-completion of this command lists only public packages from
    pypi.org and not from additional or changed python package repositories you
    may have configured on the server.
    """
    app.echo_info(f"Install package {package} ... ", nl=False)
    try:
        install_response = install_package_by_file(package_file=package)
    except FileNotFoundError as not_found_error:
        if not _looks_like_a_package(package):
            raise ClickException(
                f"{package} does not look like a package name or requirement "
                "string, and a file with this name also does not exists."
            ) from not_found_error
        install_response = install_package_by_name(package_name=package)

    # DI >= 24.1 has a combine 'output' key, before 24.1 'standardOutput' and 'errorOutput' existed
    install_output: list[str] = []
    install_output.extend(install_response.get("output", "").splitlines())
    install_output.extend(install_response.get("standardOutput", "").splitlines())
    install_output.extend(install_response.get("errorOutput", "").splitlines())
    app.echo_debug(install_output)

    update_plugin_response = update_plugins()
    app.echo_debug(f"Updated Plugins: {update_plugin_response!s}")
    update_errors = update_plugin_response.get("errors", [])
    if install_response["success"] is True and len(update_errors) == 0:
        app.echo_success("done")
        return

    # something went wrong
    app.echo_error("error")
    app.echo_error(install_output, prepend_line=True)
    for update_error in update_errors:
        app.echo_error(
            f"Error while updating the plugins of "
            f"{update_error.get('packageName')}: {update_error.get('errorMessage')} "
            f"({update_error.get('errorType')})",
            prepend_line=True,
        )
        app.echo_error(update_error.get("stackTrace"))
    sys.exit(1)


@click.command(cls=CmemcCommand, name="uninstall")
@click.argument("PACKAGE_NAME", nargs=-1, shell_complete=completion.installed_package_names)
@click.option(
    "-a",
    "--all",
    "all_",
    is_flag=True,
    help="This option removes all installed packages from the system,"
    " leaving only the pre-installed mandatory packages in the environment.",
)
@click.pass_obj
def uninstall_command(app: ApplicationContext, package_name: str, all_: bool) -> None:
    """Uninstall a python packages from the workspace.

    This command is essentially a `pip uninstall` in the remote
    python environment.
    """
    if all_:
        app.echo_info("Wiping Python environment ... ", nl=False)
        uninstall_package(package_name="--all")
        app.echo_success("done")
        app.echo_debug("Updated Plugins: " + str(update_plugins()))
        return

    if not package_name:
        raise UsageError(
            "Either give at least one package name or wipe the whole python"
            " environment with the '--all' option."
        )
    packages = list_packages()
    app.echo_debug(packages)
    for _ in package_name:
        app.echo_info(f"Uninstall package {_} ... ", nl=False)
        if _ not in [package["name"] for package in packages]:
            app.echo_error("not installed")
            app.echo_debug("Updated Plugins: " + str(update_plugins()))
            sys.exit(1)
        response = uninstall_package(package_name=_)
        output: list[str] = []
        output.extend(response.get("output", "").splitlines())
        output.extend(response.get("standardOutput", "").splitlines())
        output.extend(response.get("errorOutput", "").splitlines())
        for output_line in output:
            app.echo_debug(output_line)
        if response["success"]:
            app.echo_success("done")
        else:
            app.echo_error("error")
            app.echo_debug(response.content.decode())
    app.echo_debug("Updated Plugins: " + str(update_plugins()))


@click.command(cls=CmemcCommand, name="list")
@click.option("--raw", is_flag=True, help="Outputs raw JSON.")
@click.option(
    "--id-only",
    is_flag=True,
    help="Lists only package identifier. " "This is useful for piping the IDs into other commands.",
)
@click.option(
    "--available",
    is_flag=True,
    help="Instead of listing installed packages, this option lists installable packages"
    " from pypi.org, which are prefixed with 'cmem-plugin-' and so are most likely"
    " Corporate Memory plugin packages.",
)
@click.pass_obj
def list_command(app: ApplicationContext, raw: bool, id_only: bool, available: bool) -> None:
    """List installed python packages.

    This command is essentially a `pip list` in the remote python environment.

    It outputs a table of python package identifiers with version information.
    """
    if available:
        published_packages = get_published_packages()

        if raw:
            app.echo_info_json([asdict(_) for _ in published_packages])
            return

        if id_only:
            for _ in published_packages:
                app.echo_info(_.name)
            return

        table_published = [
            (_.name, _.version, str(_.published)[:10], _.description) for _ in published_packages
        ]
        app.echo_info_table(
            table_published,
            headers=["Name", "Version", "Published", "Description"],
            sort_column=0,
            empty_table_message="No available python packages found.",
        )
        return

    installed_packages = list_packages()
    if raw:
        app.echo_info_json(installed_packages)
        return

    if id_only:
        for package in installed_packages:
            app.echo_info(package["name"])
        return

    table_installed = [(_["name"], _["version"]) for _ in installed_packages]
    app.echo_info_table(
        table_installed,
        headers=["Name", "Version"],
        sort_column=0,
        empty_table_message="No installed python packages found."
        "Most likely, this is due to a wrong deployment.",
    )


@click.command(cls=CmemcCommand, name="list-plugins")
@click.option("--raw", is_flag=True, help="Outputs raw JSON.")
@click.option("--id-only", is_flag=True, help="Lists only plugin identifier.")
@click.option("--package-id-only", is_flag=True, help="Lists only plugin package identifier.")
@click.pass_obj
def list_plugins_command(
    app: ApplicationContext, raw: bool, id_only: bool, package_id_only: bool
) -> None:
    """List installed workspace plugins.

    This commands lists all discovered plugins.

    Note: The plugin discovery is restricted to package prefix (`cmem-`).
    """
    raw_output = list_plugins()
    try:
        plugins = raw_output["plugins"]  # DI >= 22.1.1 output
    except TypeError:
        plugins = raw_output  # DI <= 22.1 output

    if not all(plugin.get("isRegistered", True) for plugin in plugins):
        app.echo_warning(
            "Some plugins are installed but not registered. "
            "Use the 'admin workspace python reload' command to register all installed plugins."
        )

    if raw:
        app.echo_info_json(plugins)
        return

    if id_only:
        for plugin in sorted(plugins, key=lambda k: k["id"].lower()):
            app.echo_info(plugin["id"])
        return

    if package_id_only:
        package_ids = set()
        for plugin in plugins:
            package_ids.add(_get_package_id(plugin["moduleName"]))
        for package_id in sorted(package_ids):
            app.echo_info(package_id)
        return

    table = [
        (
            _["id"],
            _get_package_id(_["moduleName"]),
            _["pluginType"],
            _["label"],
        )
        for _ in sorted(plugins, key=lambda k: k["id"].lower())
    ]
    app.echo_info_table(
        table,
        headers=["ID", "Package ID", "Type", "Label"],
        sort_column=0,
        empty_table_message="No plugin plugins found. "
        "Use the `admin workspace python install` command to install python packages with plugins.",
    )
    if "error" in raw_output:
        app.echo_error(raw_output["error"])


@click.command(cls=CmemcCommand, name="open")
@click.argument(
    "PACKAGE",
    shell_complete=completion.published_package_names,
)
@click.pass_obj
def open_command(app: ApplicationContext, package: str) -> None:
    """Open a package pypi.org page in the browser.

    With this command, you can open the pypi.org page of a published package in your
    browser. From there, you can follow links, review the version history as well
    as the origin of the package, and read the provided documentation.
    """
    full_url = f"https://pypi.org/project/{package}/"
    app.echo_debug(f"Open {package}: {full_url}")
    click.launch(full_url)


@click.command(cls=CmemcCommand, name="reload")
@click.pass_obj
def reload_command(app: ApplicationContext) -> None:
    """Reload / Register all installed plugins.

    This command will register all installed plugins into the Build
    (DataIntegration) workspace.
    This command is useful, when you are installing packages
    into the Build Python environment without using the provided cmemc
    commands (e.g. by mounting a prepared filesystem in the docker container).
    """
    app.echo_info("Reloading python packages ... ", nl=False)
    update_plugin_response = update_plugins()
    app.echo_debug(f"Updated Plugins: {update_plugin_response!s}")
    update_errors = update_plugin_response.get("errors", [])
    if len(update_errors) == 0:
        app.echo_success("done")
    else:
        app.echo_error("error (use --debug for more information)")


@click.group(cls=CmemcGroup)
def python() -> CmemcGroup:  # type: ignore[empty-body]
    """List, install, or uninstall python packages.

    Python packages are used to extend the Build (DataIntegration) workspace
    with python plugins. To get a list of installed packages, execute the
    list command.

    Warning: Installing packages from unknown sources is not recommended.
    Plugins are not verified for malicious code.
    """


python.add_command(install_command)
python.add_command(uninstall_command)
python.add_command(list_command)
python.add_command(list_plugins_command)
python.add_command(open_command)
python.add_command(reload_command)



"""metrics commands for cmem command line interface."""

import click
from click import Argument, ClickException, Context, UsageError
from click.shell_completion import CompletionItem
from cmem.cmempy.api import request
from cmem.cmempy.config import get_cmem_base_uri, get_di_api_endpoint, get_dp_api_endpoint
from prometheus_client.parser import text_string_to_metric_families
from requests import HTTPError

from cmem_cmemc import completion
from cmem_cmemc.command import CmemcCommand
from cmem_cmemc.command_group import CmemcGroup
from cmem_cmemc.context import ApplicationContext
from cmem_cmemc.object_list import (
    DirectValuePropertyFilter,
    ObjectList,
    compare_regex,
    compare_str_equality,
)
from cmem_cmemc.utils import (
    metric_get_labels,
    struct_to_table,
)


def get_all_metrics(ctx: click.Context) -> list[dict]:
    """Get metrics data for object list"""
    known_metrics_urls: list[tuple[str, str]] = [
        ("explore", get_dp_api_endpoint() + "/actuator/prometheus"),
        ("build", get_di_api_endpoint() + "/metrics"),
        ("store", get_dp_api_endpoint() + "/actuator/proxy/graphdb/metrics/structures"),
        ("store", get_dp_api_endpoint() + "/actuator/proxy/graphdb/metrics/infrastructure"),
        ("store", get_dp_api_endpoint() + "/actuator/proxy/graphdb/metrics/cluster"),
        ("store", get_dp_api_endpoint() + "/actuator/proxy/graphdb/metrics/repository/cmem"),
    ]
    app: ApplicationContext = ctx.obj
    all_metrics: list[dict] = []
    for job, url in known_metrics_urls:
        try:
            families = text_string_to_metric_families(request(url).text)
        except HTTPError as error:
            if app is not None:
                # when in completion mode, obj is not set :-(
                app.echo_debug(str(error))
            continue
        for family in families:
            documentation = family.documentation
            if documentation == "":
                documentation = f"No documentation available for {job}:{family.name}"
            new_metric = {
                "id": f"{job}:{family.name}",
                "job": job,
                "name": family.name,
                "type": family.type,
                "documentation": documentation,
                "labels": metric_get_labels(family),
                "samples": family.samples,
                # "object": family
            }
            if family.type not in ("unknown", "histogram"):
                all_metrics.append(new_metric)
    return all_metrics


metrics_list = ObjectList(
    name="metrics",
    get_objects=get_all_metrics,
    filters=[
        DirectValuePropertyFilter(
            name="job",
            description="Filter metrics by job ID.",
            property_key="job",
        ),
        DirectValuePropertyFilter(
            name="name",
            description="Filter metrics by regex matching the name.",
            property_key="name",
            compare=compare_regex,
            fixed_completion=[],
        ),
        DirectValuePropertyFilter(
            name="type",
            description="Filter metrics by type.",
            property_key="type",
        ),
        DirectValuePropertyFilter(
            name="id",
            description="Filter metrics by ID.",
            property_key="id",
            compare=compare_str_equality,
        ),
    ],
)


def _complete_metrics_id(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:  # noqa: ARG001
    """Prepare a list of metric identifier."""
    ApplicationContext.set_connection_from_params(ctx.find_root().params)
    candidates = [(_["id"], _["documentation"]) for _ in metrics_list.apply_filters(ctx=ctx)]
    return completion.finalize_completion(candidates=candidates, incomplete=incomplete)


def _complete_metric_label_filter(
    ctx: Context,
    param: Argument,  # noqa: ARG001
    incomplete: str,
) -> list[CompletionItem]:
    """Prepare a list of label name or values"""
    ApplicationContext.set_connection_from_params(ctx.find_root().params)
    args = completion.get_completion_args(incomplete)
    incomplete = incomplete.lower()
    options: list[str] = []
    try:
        metric_id = ctx.args[0]
        metric = metrics_list.apply_filters(ctx=ctx, filter_=[("id", metric_id)])[0]
    except IndexError:
        # means: --filter is used before we have a metrics ID
        return []
    labels = metric["labels"]
    if args[len(args) - 1] in "--filter":
        # we are in the name position
        options = list(labels.keys())
    if args[len(args) - 2] in "--filter":
        label_name = args[len(args) - 1]
        # we are in the value position
        options = labels[label_name]
    return completion.finalize_completion(candidates=options, incomplete=incomplete)


def _filter_samples(family: dict, label_filter: tuple[tuple[str, str], ...]) -> list:
    """Filter samples by labels."""
    family_name = family.get("name")
    all_samples: list[list] = family["samples"]
    if not label_filter:
        return all_samples
    labels: dict[str, list[str]] = family["labels"]
    samples = []
    for sample in all_samples:
        matching_labels = 0
        sample_labels = sample[1]
        for name, value in label_filter:
            if name not in labels:
                raise ClickException(
                    f"The metric '{family_name}' does " f"not have a label named '{name}'."
                )
            if value not in labels[name]:
                raise ClickException(
                    f"The metric '{family_name}' does "
                    f"not have a label '{name}' with the value '{value}'."
                )
            if name in sample_labels and sample_labels[name] == value:
                matching_labels += 1
        if matching_labels == len(label_filter):
            # all filter match
            samples.append(sample)
    return samples


# pylint: disable-msg=too-many-arguments
@click.command(cls=CmemcCommand, name="get")
@click.argument("metric_id", required=True, type=click.STRING, shell_complete=_complete_metrics_id)
@click.option(
    "--filter",
    "label_filter",
    type=(str, str),
    shell_complete=_complete_metric_label_filter,
    multiple=True,
    help="A set of label name/value pairs in order to filter the samples "
    "of the requested metric family. Each metric has a different set "
    "of labels with different values. "
    "In order to get a list of possible label names and values, use "
    "the command without this option. The label names are then shown "
    "as column headers and label values as cell values of this column.",
)
@click.option(
    "--enforce-table",
    is_flag=True,
    help="A single sample value will be returned as plain text instead "
    "of the normal table. This allows for more easy integration "
    "with scripts. This flag enforces the use of tabular output, "
    "even for single row tables.",
)
@click.option("--raw", is_flag=True, help="Outputs raw prometheus sample classes.")
@click.pass_context
def get_command(
    ctx: click.Context,
    metric_id: str,
    label_filter: tuple[tuple[str, str], ...],
    raw: bool,
    enforce_table: bool,
) -> None:
    """Get sample data of a metric.

    A metric of a specific job is identified by a metric ID. Possible
    metric IDs of a job can be retrieved with the `metrics list`
    command. A metric can contain multiple samples.
    These samples are distinguished by labels (name and value).
    """
    app = ctx.obj
    data = metrics_list.apply_filters(ctx=ctx, filter_=[("id", metric_id)])
    if len(data) == 0:
        raise UsageError(
            f"No metric with ID '{metric_id}' found. "
            "Use the `metrics list` command to list all metrics."
        )
    if len(data) > 1:
        raise UsageError("Unknown Error - More than one metric with ID '{metric_id}' found.")
    metric = data[0]

    samples = _filter_samples(metric, label_filter)
    if raw:
        app.echo_info_json(samples)
        return

    if len(samples) == 0:
        raise ClickException(
            "No data - the given label combination filtered out "
            f"all available samples of the metric {metric_id}."
        )

    if len(samples) == 1 and enforce_table is not True:
        app.echo_info(str(samples[0].value))
        return

    label_dict = metric["labels"]
    table = []
    for sample in samples:
        row = [sample.labels[key] for key in sorted(label_dict.keys())]
        table.append(row)
        row.append(str(sample.value))
    headers = sorted(label_dict.keys())
    headers.append("value")
    app.echo_info_table(table, headers=headers, sort_column=0)


@click.command(cls=CmemcCommand, name="inspect")
@click.argument("metric_id", required=True, type=click.STRING, shell_complete=_complete_metrics_id)
@click.option("--raw", is_flag=True, help="Outputs raw JSON of the table data.")
@click.pass_context
def inspect_command(ctx: Context, metric_id: str, raw: bool) -> None:
    """Inspect a metric.

    This command outputs the data of a metric.
    The first table includes basic metadata about the metric.
    The second table includes sample labels and values.
    """
    app = ctx.obj
    data = metrics_list.apply_filters(ctx=ctx, filter_=[("id", metric_id)])
    if len(data) == 0:
        raise UsageError(
            f"No metric with ID '{metric_id}' found. "
            "Use the `metrics list` command to list all metrics."
        )
    if len(data) > 1:
        raise UsageError("Unknown Error - More than one metric with ID '{metric_id}' found.")
    if raw:
        app.echo_info_json(data)
        return
    app.echo_info_table(struct_to_table(data), headers=["Key", "Value"], sort_column=0)


@click.command(cls=CmemcCommand, name="list")
@click.option(
    "--filter",
    "filter_",
    type=(str, str),
    multiple=True,
    help=metrics_list.get_filter_help_text(),
    shell_complete=metrics_list.complete_values,
)
@click.option(
    "--id-only",
    is_flag=True,
    help="Lists metric identifier only. " "This is useful for piping the IDs into other commands.",
)
@click.option(
    "--raw", is_flag=True, help="Outputs (sorted) JSON dict, parsed from the metrics API output."
)
@click.pass_context
def list_command(
    ctx: click.Context, filter_: tuple[tuple[str, str]], id_only: bool, raw: bool
) -> None:
    """List metrics for a specific job.

    For each metric, the output table shows the metric ID,
    the type of the metric, a count of how many labels (label names)
    are describing the samples (L) and a count of how many samples are
    currently available for a metric (S).
    """
    app = ctx.obj
    data = metrics_list.apply_filters(ctx=ctx, filter_=filter_)
    if raw:
        app.echo_info_json(data)
        return
    if id_only:
        for _ in sorted([_.get("id") for _ in data]):
            app.echo_info(_)
        return

    table = [
        [
            _.get("id"),
            _.get("type"),
            len(_.get("labels")),
            len(_.get("samples")),
            _.get("documentation"),
        ]
        for _ in data
    ]
    app.echo_info_table(
        table,
        headers=["ID", "Type", "L", "S", "Documentation"],
        sort_column=0,
        caption=f"{len(table)} metrics families of {get_cmem_base_uri()}",
        empty_table_message="No metrics families available.",
    )


@click.group(cls=CmemcGroup)
def metrics() -> CmemcGroup:  # type: ignore[empty-body]
    """List and get metrics.

    This command group consists of commands for reading and listing
    internal monitoring metrics of eccenca Corporate Memory. A
    deployment consists of multiple jobs (e.g. DP, DI), which provide
    multiple metric families for an endpoint.

    Each metric family can consist of different samples identified by
    labels with a name and a value (dimensions). A metric has a specific
    type (counter, gauge, summary and histogram) and additional metadata.

    Please have a look at https://prometheus.io/docs/concepts/data_model/
    for further details.
    """


metrics.add_command(get_command)
metrics.add_command(inspect_command)
metrics.add_command(list_command)



"""Bootstrap Data migration recipe"""

from typing import ClassVar

from cmem.cmempy.dp.admin import import_bootstrap_data
from cmem.cmempy.health import get_complete_status_info

from cmem_cmemc.migrations.abc import MigrationRecipe, components


class MigrateBootstrapData(MigrationRecipe):
    """Update Bootstrap Data to current version"""

    id = "bootstrap-data"
    description = "Migrate bootstrap system data (same as admin store bootstrap --import)"
    component: components = "explore"
    first_version = "20.12"
    tags: ClassVar[list[str]] = ["system", "shapes"]

    def is_applicable(self) -> bool:
        """Test if the recipe can be applied."""
        status_info = get_complete_status_info()
        shapes_status = status_info["shapes"]["version"]
        return bool(shapes_status != status_info["explore"]["version"])

    def apply(self) -> None:
        """Apply the recipe to the current version."""
        import_bootstrap_data()



"""Graph imports command"""

import os

import click
from click import Argument, ClickException, Context, UsageError
from click.shell_completion import CompletionItem
from cmem.cmempy.dp.proxy.graph import get_graph_import_tree
from cmem.cmempy.queries import SparqlQuery
from treelib import Tree

from cmem_cmemc import completion
from cmem_cmemc.command import CmemcCommand
from cmem_cmemc.command_group import CmemcGroup
from cmem_cmemc.completion import escape_colon
from cmem_cmemc.constants import UNKNOWN_GRAPH_ERROR
from cmem_cmemc.context import ApplicationContext
from cmem_cmemc.object_list import DirectValuePropertyFilter, ObjectList
from cmem_cmemc.string_processor import GraphLink
from cmem_cmemc.title_helper import TitleHelper
from cmem_cmemc.utils import get_graphs_as_dict, tuple_to_list

GRAPH_IMPORTS_LIST_SPARQL = """
PREFIX owl: <http://www.w3.org/2002/07/owl#>

SELECT ?from_graph ?to_graph
WHERE
{
  GRAPH ?from_graph {
    ?from_graph owl:imports ?to_graph
  }
}
"""

GRAPH_IMPORTS_CREATE_SPARQL = """
PREFIX owl: <http://www.w3.org/2002/07/owl#>

INSERT DATA {
  GRAPH <{{from_graph}}> {
    <{{from_graph}}> owl:imports <{{to_graph}}> .
  }
}
"""

GRAPH_IMPORTS_DELETE_SPARQL = """
PREFIX owl: <http://www.w3.org/2002/07/owl#>

DELETE DATA {
  GRAPH <{{from_graph}}> {
    <{{from_graph}}> owl:imports <{{to_graph}}> .
  }
}
"""


def _prepare_tree_output_id_only(iris: list[str], graphs: dict) -> str:
    """Prepare a sorted, de-duplicated IRI list of graph imports."""
    output_iris = []
    for iri in iris:
        # get response for one requested graph
        api_response = get_graph_import_tree(iri)

        # add all imported IRIs to the IRI list
        # add the requested graph as well
        output_iris.append(iri)
        for top_graph in api_response["tree"]:
            output_iris.append(top_graph)
            for sub_graph in api_response["tree"][top_graph]:
                output_iris.append(sub_graph)  # noqa: PERF402

    # prepare a sorted, de-duplicated IRI list of existing graphs
    # and create a line-by-line output of it
    output_iris = sorted(set(output_iris), key=lambda x: x.lower())
    filtered_iris = [iri for iri in output_iris if iri in graphs]
    return "\n".join(filtered_iris[0:]) + "\n"


def _create_node_label(iri: str, graphs: dict) -> str:
    """Create a label for a node in the tree."""
    if iri not in graphs:
        return "[missing: " + iri + "]"
    title = graphs[iri]["label"]["title"]
    return f"{title} -- {iri}"


def _add_tree_nodes_recursive(tree: Tree, structure: dict, iri: str, graphs: dict) -> Tree:
    """Add all child nodes of iri from structure to tree.

    Call recursively until no child node can be used as parent anymore.

    Args:
    ----
        tree: the graph where to add the nodes
        structure: the result dict of get_graph_import_tree()
        iri: The IRI of the parent
        graphs: the result of get_graphs()

    Returns:
    -------
        the new treelib.Tree object with the additional nodes

    """
    if not tree.contains(iri):
        tree.create_node(tag=_create_node_label(iri, graphs), identifier=iri)
    if iri not in structure:
        return tree
    for child in structure[iri]:
        tree.create_node(tag=_create_node_label(child, graphs), identifier=child, parent=iri)
    for child in structure[iri]:
        if child in structure:
            tree = _add_tree_nodes_recursive(tree, structure, child, graphs)
    return tree


def _add_ignored_nodes(tree: Tree, structure: dict) -> Tree:
    """Add all child nodes as ignored nodes.

    Args:
    ----
        tree: the graph where to add the nodes
        structure: the result dict of get_graph_import_tree()

    Returns:
    -------
        the new treelib.Tree object with the additional nodes

    """
    if len(structure.keys()) > 0:
        for parent in structure:
            for children in structure[parent]:
                tree.create_node(tag="[ignored: " + children + "]", parent=parent)
    return tree


def get_imports_list(ctx: click.Context) -> list[dict[str, str]]:  # noqa: ARG001
    """Get the import list"""
    list_query = SparqlQuery(text=GRAPH_IMPORTS_LIST_SPARQL)
    result = list_query.get_json_results()
    return [
        {"from_graph": _["from_graph"]["value"], "to_graph": _["to_graph"]["value"]}
        for _ in result["results"]["bindings"]
    ]


@click.command(cls=CmemcCommand, name="tree")
@click.option("-a", "--all", "all_", is_flag=True, help="Show tree of all (readable) graphs.")
@click.option("--raw", is_flag=True, help="Outputs raw JSON of the graph importTree API response.")
@click.option(
    "--id-only",
    is_flag=True,
    help="Lists only graph identifier (IRIs) and no labels or other "
    "metadata. This is useful for piping the IRIs into other commands. "
    "The output with this option is a sorted, flat, de-duplicated list "
    "of existing graphs.",
)
@click.argument(
    "iris",
    nargs=-1,
    type=click.STRING,
    shell_complete=completion.graph_uris,
    callback=tuple_to_list,
)
@click.pass_obj
def tree_command(
    app: ApplicationContext, all_: bool, raw: bool, id_only: bool, iris: list[str]
) -> None:
    """Show graph tree(s) of the imports statement hierarchy.

    You can output one or more trees of the import hierarchy.

    Imported graphs which do not exist are shown as `[missing: IRI]`.
    Imported graphs which will result in an import cycle are shown as
    `[ignored: IRI]`.
    Each graph is shown with label and IRI.
    """
    graphs = get_graphs_as_dict()
    if not iris and not all_:
        raise UsageError(
            "Either specify at least one graph IRI or use the "
            "--all option to show the owl:imports tree of all graphs."
        )
    if all_:
        iris = [str(_) for _ in graphs]

    for iri in iris:
        if iri not in graphs:
            raise ClickException(UNKNOWN_GRAPH_ERROR.format(iri))

    iris = sorted(iris, key=lambda x: graphs[x]["label"]["title"].lower())

    if raw:
        for iri in iris:
            # direct output of the response for one requested graph
            app.echo_info_json(get_graph_import_tree(iri))
        return

    if id_only:
        app.echo_result(_prepare_tree_output_id_only(iris, graphs), nl=False)
        return

    # normal execution
    output = ""
    for iri in iris:
        # get response for on requested graph
        api_response = get_graph_import_tree(iri)

        tree = _add_tree_nodes_recursive(Tree(), api_response["tree"], iri, graphs)
        tree = _add_ignored_nodes(tree, api_response["ignored"])

        # strip empty lines from the tree.show output
        output += os.linesep.join(
            [
                line
                for line in tree.show(key=lambda x: x.tag.lower(), stdout=False).splitlines()  # type: ignore[arg-type, return-value]
                if line.strip()
            ]
        )
        output += "\n"
    # result output
    app.echo_result(output, nl=False)


graph_imports_list = ObjectList(
    name="imports",
    get_objects=get_imports_list,
    filters=[
        DirectValuePropertyFilter(
            name="from-graph",
            description="List only matches from graph",
            property_key="from_graph",
            title_helper=TitleHelper(),
        ),
        DirectValuePropertyFilter(
            name="to-graph",
            description="List only matches to graph",
            property_key="to_graph",
            title_helper=TitleHelper(),
        ),
    ],
)


@click.command(cls=CmemcCommand, name="list")
@click.option("--raw", is_flag=True, help="Outputs raw JSON response.")
@click.option(
    "--filter",
    "filter_",
    type=(str, str),
    help=graph_imports_list.get_filter_help_text(),
    shell_complete=graph_imports_list.complete_values,
)
@click.pass_context
def list_command(ctx: Context, raw: bool, filter_: tuple[str, str]) -> None:
    """List accessible graph imports statements.

    Graphs are identified by an IRI. Statement imports are managed by
    creating owl:imports statements such as "FROM_GRAPH owl:imports TO_GRAPH"
    in the FROM_GRAPH. All statements in the TO_GRAPH are then available
    in the FROM_GRAPH.
    """
    app: ApplicationContext = ctx.obj
    filters_to_apply = []
    if filter_:
        filters_to_apply.append(filter_)
    imports = graph_imports_list.apply_filters(ctx=ctx, filter_=filters_to_apply)

    if raw:
        app.echo_info_json(imports)
        return

    table = []
    graphs = get_graphs_as_dict()
    for _ in imports:
        from_graph = _["from_graph"]
        to_graph = _["to_graph"]
        if to_graph not in graphs:
            to_graph = rf"\[missing: {to_graph}]"
        table.append([from_graph, to_graph])

    app.echo_info_table(
        table,
        headers=["From graph", "To graph"],
        sort_column=0,
        cell_processing={0: GraphLink(), 1: GraphLink()},
        empty_table_message="No imports found. "
        "You can use the `graph imports create` command to create a graph import.",
    )


def _validate_graphs(from_graph: str | None, to_graph: str | None) -> None:
    graphs = get_graphs_as_dict(writeable=True, readonly=True)
    if from_graph and from_graph not in graphs:
        raise click.UsageError(f"From graph {from_graph} not found.")

    if to_graph and to_graph not in graphs:
        raise click.UsageError(f"To graph {to_graph} not found.")


def _from_graph_uris(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Provide auto completion items for delete command from-graph argument"""
    imports = get_imports_list(ctx)
    from_graphs = {escape_colon(_["from_graph"]) for _ in imports}
    return [
        _
        for _ in completion.graph_uris(ctx=ctx, param=param, incomplete=incomplete)
        if _.value in from_graphs
    ]


def _to_graph_uris(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Provide auto completion items for create/delete command to-graph argument"""
    from_graph = ctx.params["from_graph"]
    imports = graph_imports_list.apply_filters(ctx=ctx, filter_=[("from-graph", from_graph)])
    to_graphs = {escape_colon(_["to_graph"]) for _ in imports}
    command = ctx.command.name
    return [
        _
        for _ in completion.graph_uris(ctx=ctx, param=param, incomplete=incomplete)
        if (command == "delete" and _.value in to_graphs)
        or (
            command == "create" and _.value not in to_graphs and _.value != escape_colon(from_graph)
        )
    ]


@click.command(cls=CmemcCommand, name="create")
@click.argument("from_graph", type=str, shell_complete=completion.graph_uris)
@click.argument("to_graph", type=str, shell_complete=_to_graph_uris)
@click.pass_context
def create_command(ctx: Context, from_graph: str, to_graph: str) -> None:
    """Add statement to import a TO_GRAPH into a FROM_GRAPH.

    Graphs are identified by an IRI. Statement imports are managed by
    creating owl:imports statements such as "FROM_GRAPH owl:imports TO_GRAPH"
    in the FROM_GRAPH. All statements in the TO_GRAPH are then available
    in the FROM_GRAPH.

    Note: The get a list of existing graphs, execute the `graph list` command or
    use tab-completion.
    """
    app: ApplicationContext = ctx.obj
    _validate_graphs(from_graph, to_graph)
    if from_graph == to_graph:
        raise click.UsageError("From graph and to graph cannot be the same.")

    imports = graph_imports_list.apply_filters(
        ctx=ctx, filter_=[("from-graph", from_graph), ("to-graph", to_graph)]
    )
    if imports:
        raise click.UsageError("Import combination already exists.")
    app.echo_info(f"Creating graph import from {from_graph} to {to_graph} ... ", nl=False)

    create_query = SparqlQuery(
        text=GRAPH_IMPORTS_CREATE_SPARQL,
        query_type="UPDATE",
    )
    create_query.get_results(
        placeholder={
            "from_graph": from_graph,
            "to_graph": to_graph,
        }
    )
    app.echo_success("done")


@click.command(cls=CmemcCommand, name="delete")
@click.argument("from_graph", type=str, shell_complete=_from_graph_uris)
@click.argument("to_graph", type=str, shell_complete=_to_graph_uris)
@click.pass_context
def delete_command(ctx: Context, from_graph: str, to_graph: str) -> None:
    """Delete statement to import a TO_GRAPH into a FROM_GRAPH.

    Graphs are identified by an IRI. Statement imports are managed by
    creating owl:imports statements such as "FROM_GRAPH owl:imports TO_GRAPH"
    in the FROM_GRAPH. All statements in the TO_GRAPH are then available
    in the FROM_GRAPH.

    Note: The get a list of existing graph imports, execute the
    `graph imports list` command or use tab-completion.
    """
    app: ApplicationContext = ctx.obj
    _validate_graphs(from_graph, None)
    imports = graph_imports_list.apply_filters(
        ctx=ctx, filter_=[("from-graph", from_graph), ("to-graph", to_graph)]
    )
    if not imports:
        raise click.UsageError("Import combination does not exists.")
    app.echo_info(f"Deleting graph import from {from_graph} to {to_graph} ... ", nl=False)

    delete_query = SparqlQuery(
        text=GRAPH_IMPORTS_DELETE_SPARQL,
        query_type="UPDATE",
    )
    delete_query.get_results(
        placeholder={
            "from_graph": from_graph,
            "to_graph": to_graph,
        }
    )
    app.echo_success("done")


@click.group(cls=CmemcGroup, name="imports")
def imports_group() -> CmemcGroup:  # type: ignore[empty-body]
    """List, create, delete and show graph imports.

    Graphs are identified by an IRI. Statement imports are managed by
    creating owl:imports statements such as "FROM_GRAPH owl:imports TO_GRAPH"
    in the FROM_GRAPH. All statements in the TO_GRAPH are then available
    in the FROM_GRAPH.

    Note: The get a list of existing graphs,
    execute the `graph list` command or use tab-completion.
    """


imports_group.add_command(tree_command)
imports_group.add_command(list_command)
imports_group.add_command(create_command)
imports_group.add_command(delete_command)



"""Migration: Chart and Workflow Property Shapes to Widget Integration"""

from typing import ClassVar

from cmem_cmemc.migrations.abc import MigrationRecipe, components


class ChartsOnNodeShapesToToWidgetIntegrations(MigrationRecipe):
    """24.3 Migrate Chart on NodeShapes to Widget Integrations"""

    id = "charts-on-nshapes-24.3"
    description = "Migrate Charts on Node Shapes to Widget Integrations"
    component: components = "explore"
    first_version = "24.3"
    tags: ClassVar[list[str]] = ["shapes", "user"]
    check_query = """{{DEFAULT_PREFIXES}}
SELECT *
WHERE {
  GRAPH ?maybeOtherGraph {
    ?maybeOtherGraph a shui:ShapeCatalog .
    ?nodeShape a sh:NodeShape .
  }
  GRAPH ?shapeCatalog {
    ?shapeCatalog a shui:ShapeCatalog .
    ?nodeShape shui:provideChartVisualization ?shuiChart .
  }
}
    """
    move_query = """{{DEFAULT_PREFIXES}}
DELETE {
  GRAPH ?shapeCatalog {
    ?nodeShape shui:provideChartVisualization ?shuiChart .
  }
}
INSERT {
  GRAPH ?shapeCatalog {
    ?newWidgetIntegration a shui:WidgetIntegration ;
      shui:WidgetIntegration_widget ?shuiChart ;
      rdfs:label ?name ;
      rdfs:comment ?description ;
      sh:order -1000 .
    ?nodeShape shui:WidgetIntegration_integrate ?newWidgetIntegration .
  }
}
WHERE {
  GRAPH ?maybeOtherGraph {
    ?maybeOtherGraph a shui:ShapeCatalog .
    ?nodeShape a sh:NodeShape .
  }
  GRAPH ?shapeCatalog {
    ?shapeCatalog a shui:ShapeCatalog .
    ?nodeShape shui:provideChartVisualization ?shuiChart .
  }
  OPTIONAL { ?shuiChart rdfs:label ?name }
  OPTIONAL { ?shuiChart rdfs:comment ?description }
  BIND (IRI(CONCAT(STR(?shuiChart), "_WidgetIntegration")) as ?newWidgetIntegration)
}
    """

    def is_applicable(self) -> bool:
        """Test if the recipe can be applied."""
        node_shapes_with_charts = self._select(self.check_query)
        return len(node_shapes_with_charts) > 0

    def apply(self) -> None:
        """Apply the recipe to the current version."""
        self._update(self.move_query)


class ChartsOnPropertyShapesToWidgetIntegrations(MigrationRecipe):
    """24.3 Migrate Chart Property Shapes to Widget Integrations"""

    id = "charts-on-pshapes-24.3"
    description = "Migrate Charts on Property Shapes to Widget Integrations"
    component: components = "explore"
    first_version = "24.3"
    tags: ClassVar[list[str]] = ["shapes", "user"]
    check_query = """{{DEFAULT_PREFIXES}}
SELECT *
WHERE {
  GRAPH ?shapeCatalog {
    ?shapeCatalog a shui:ShapeCatalog .
    ?propertyShape a sh:PropertyShape ;
      shui:provideChartVisualization ?shuiChart .
  }
}
    """
    move_query = """{{DEFAULT_PREFIXES}}
DELETE {
  GRAPH ?shapeCatalog {
    ?propertyShape ?p ?o .
    ?nodeShape sh:property ?propertyShape .
  }
}
INSERT {
  GRAPH ?shapeCatalog {
    ?propertyShape a shui:WidgetIntegration ;
      shui:WidgetIntegration_widget ?shuiChart ;
      rdfs:label ?name ;
      rdfs:comment ?description ;
      sh:order ?order ;
      shui:WidgetIntegration_group ?group .
    ?nodeShape shui:WidgetIntegration_integrate ?propertyShape .
  }
}
WHERE {
  GRAPH ?shapeCatalog {
    ?shapeCatalog a shui:ShapeCatalog .
    ?propertyShape a sh:PropertyShape ;
      shui:provideChartVisualization ?shuiChart .
    ?propertyShape ?p ?o .
    OPTIONAL {?propertyShape sh:name ?name }
    OPTIONAL {?propertyShape sh:description ?description }
    OPTIONAL {?propertyShape sh:order ?order }
    OPTIONAL {?propertyShape sh:group ?group }
    OPTIONAL {?nodeShape sh:property ?propertyShape }
  }
}
    """

    def is_applicable(self) -> bool:
        """Test if the recipe can be applied."""
        chart_property_shapes = self._select(self.check_query)
        return len(chart_property_shapes) > 0

    def apply(self) -> None:
        """Apply the recipe to the current version."""
        self._update(self.move_query)


class WorkflowTriggerPropertyShapesToWidgetIntegrations(MigrationRecipe):
    """24.3 Migrate Workflow Trigger Property Shapes to Widget Integrations"""

    id = "trigger-on-pshapes-24.3"
    description = "Migrate Workflow Trigger on Property Shapes to Widget Integrations"
    component: components = "explore"
    first_version = "24.3"
    tags: ClassVar[list[str]] = ["shapes", "user"]
    check_query = """{{DEFAULT_PREFIXES}}
SELECT *
WHERE {
  GRAPH ?shapeCatalog {
    ?shapeCatalog a shui:ShapeCatalog .
    ?propertyShape a sh:PropertyShape ;
      shui:provideWorkflowTrigger ?shuiChart .
  }
}
    """
    move_query = """{{DEFAULT_PREFIXES}}
DELETE {
  GRAPH ?shapeCatalog {
    ?propertyShape ?p ?o .
    ?nodeShape sh:property ?propertyShape .
  }
}
INSERT {
  GRAPH ?shapeCatalog {
    ?propertyShape a shui:WidgetIntegration ;
      shui:WidgetIntegration_widget ?workflowTrigger ;
      rdfs:label ?name ;
      rdfs:comment ?description ;
      sh:order ?order ;
      shui:WidgetIntegration_group ?group .
    ?nodeShape shui:WidgetIntegration_integrate ?propertyShape .
  }
}
WHERE {
  GRAPH ?shapeCatalog {
    ?shapeCatalog a shui:ShapeCatalog .
    ?propertyShape a sh:PropertyShape ;
      shui:provideWorkflowTrigger ?workflowTrigger .
    ?propertyShape ?p ?o .
    OPTIONAL {?propertyShape sh:name ?name }
    OPTIONAL {?propertyShape sh:description ?description }
    OPTIONAL {?propertyShape sh:order ?order }
    OPTIONAL {?propertyShape sh:group ?group }
    OPTIONAL {?nodeShape sh:property ?propertyShape }
  }
}
    """

    def is_applicable(self) -> bool:
        """Test if the recipe can be applied."""
        chart_property_shapes = self._select(self.check_query)
        return len(chart_property_shapes) > 0

    def apply(self) -> None:
        """Apply the recipe to the current version."""
        self._update(self.move_query)


class TableReportPropertyShapesToWidgetIntegrations(MigrationRecipe):
    """24.3 Migrate Table Report Property Shapes to Widget Integrations"""

    id = "table-reports-on-pshapes-24.3"
    description = (
        "Migrate Table Reports (which use shui:null) on Property Shapes to Widget Integrations"
    )
    component: components = "explore"
    first_version = "24.3"
    tags: ClassVar[list[str]] = ["shapes", "user"]
    check_query = """{{DEFAULT_PREFIXES}}
SELECT ?propertyShape
WHERE {
  GRAPH ?shapeCatalog {
    ?shapeCatalog a shui:ShapeCatalog .
    ?propertyShape a sh:PropertyShape ;
      shui:valueQuery ?valueQuery ;
      sh:path shui:null .
    FILTER NOT EXISTS { ?propertyShape shui:isSystemResource true }
  }
}
    """
    move_query = """{{DEFAULT_PREFIXES}}
DELETE {
  GRAPH ?shapeCatalog {
    ?propertyShape ?p ?o .
    ?nodeShape sh:property ?propertyShape .
  }
}
INSERT {
  GRAPH ?shapeCatalog {
    ?propertyShape a shui:WidgetIntegration ;
      shui:WidgetIntegration_widget ?newTableReport ;
      rdfs:label ?name ;
      rdfs:comment ?description ;
      sh:order ?order ;
      shui:WidgetIntegration_group ?group .
    ?nodeShape shui:WidgetIntegration_integrate ?propertyShape .

    ?newTableReport a shui:TableReport ;
      rdfs:label ?newTableReportLabel ;
      shui:TableReport_hideFooter ?hideFooter ;
      shui:TableReport_hideHeader ?hideHeader ;
      shui:TableReport_query ?valueQuery .
  }
}
WHERE {
  GRAPH ?shapeCatalog {
    ?shapeCatalog a shui:ShapeCatalog .
    ?propertyShape a sh:PropertyShape ;
      sh:path shui:null ;
      shui:valueQuery ?valueQuery .
    FILTER NOT EXISTS { ?propertyShape shui:isSystemResource true }
    ?propertyShape ?p ?o .
    OPTIONAL {?propertyShape sh:name ?name }
    OPTIONAL {?propertyShape sh:description ?description }
    OPTIONAL {?propertyShape sh:order ?order }
    OPTIONAL {?propertyShape sh:group ?group }
    OPTIONAL {?propertyShape sh:name ?name }
    OPTIONAL {?propertyShape shui:valueQueryHideHeader ?hideHeaderValue }
    OPTIONAL {?propertyShape shui:valueQueryHideFooter ?hideFooterValue }
    OPTIONAL {?nodeShape sh:property ?propertyShape }
    BIND ( IRI(CONCAT(STR(?propertyShape), "-TableReport")) AS ?newTableReport )
    BIND ( STRLANG(CONCAT("Table Report: ", ?name), lang(?name)) AS ?newTableReportLabel )
    BIND ( COALESCE(?hideHeaderValue, false) as ?hideHeader)
    BIND ( COALESCE(?hideFooterValue, false) as ?hideFooter)
  }
}
    """

    def is_applicable(self) -> bool:
        """Test if the recipe can be applied."""
        table_report_property_shapes = self._select(self.check_query)
        return len(table_report_property_shapes) > 0

    def apply(self) -> None:
        """Apply the recipe to the current version."""
        self._update(self.move_query)



"""Config parser module with a custom configuration parser."""

import configparser


class PureSectionConfigParser(configparser.RawConfigParser):
    """A configuration parser that retrieves section data without defaults."""

    _defaults: dict

    def options(self, section: str) -> list:
        """Return a list of option names for the given section, excluding defaults.

        This method temporarily removes default options to ensure only explicitly
        defined options in the given section are returned.

        Args:
            section (str): The section name.

        Returns:
            list: A list of option names in the specified section.

        """
        _d = self._defaults.copy()
        try:
            self._defaults.clear()
            return super().options(section)
        finally:
            self._defaults.update(_d)

    def optionxform(self, option: str) -> str:
        """Preserve the case of option names.

        By default, `optionxform` in `RawConfigParser` converts option names to lowercase.
        This overridden method ensures that option names remain unchanged.

        Args:
            option (str): The option name.

        Returns:
            str: The unmodified option name.

        """
        return option



"""Migration: Chart and Workflow Property Shapes to Widget Integration"""

from typing import ClassVar

from cmem_cmemc.migrations.abc import MigrationRecipe, components


class SparqlDatatypesToXsdString(MigrationRecipe):
    """24.2 Migrate shui:sparqlQuery|sparqlUpdate datatype literals to xsd:string literals"""

    id = "sparql-query-datatypes-24.2"
    description = "Migrate shui SPARQL datatype literals to xsd:string literals"
    component: components = "explore"
    first_version = "24.2"
    tags: ClassVar[list[str]] = ["shapes", "user"]
    check_query = """{{DEFAULT_PREFIXES}}
SELECT DISTINCT ?query
WHERE {
  GRAPH ?shapeOrQueryGraph {
    ?query shui:queryText ?text .
    FILTER ( datatype(?text) IN (shui:sparqlQuery, shui:sparqlUpdate) )
  }
}
    """
    move_query = """{{DEFAULT_PREFIXES}}
DELETE {
  GRAPH ?shapeOrQueryGraph {
    ?query shui:queryText ?oldLiteral .
  }
}
INSERT {
  GRAPH ?shapeOrQueryGraph {
    ?query shui:queryText ?newLiteral .
  }
}
# SELECT DISTINCT ?query
WHERE {
  GRAPH ?shapeOrQueryGraph {
    ?query shui:queryText ?oldLiteral .
    FILTER ( datatype(?oldLiteral) IN (shui:sparqlQuery, shui:sparqlUpdate) )
    BIND (STRDT(STR(?oldLiteral), xsd:string) AS ?newLiteral)
  }
}
    """

    def is_applicable(self) -> bool:
        """Test if the recipe can be applied."""
        queries_with_old_literals = self._select(self.check_query)
        return len(queries_with_old_literals) > 0

    def apply(self) -> None:
        """Apply the recipe to the current version."""
        self._update(self.move_query)



"""Client module to handle Path API calls."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Generator


class StoragePath(ABC):
    """Storage path interface."""

    def __init__(self, path: str):
        self.path = path

    @property
    @abstractmethod
    def suffix(self) -> str:
        """Return the suffix of the path."""

    @property
    @abstractmethod
    def name(self) -> str:
        """Return the file name."""

    @property
    @abstractmethod
    def parent(self) -> StoragePath:
        """The logical parent of the path."""

    @abstractmethod
    def is_dir(self) -> bool:
        """Check if the path is a directory."""

    @abstractmethod
    def is_file(self) -> bool:
        """Check if the path is a file."""

    @abstractmethod
    def exists(self) -> bool:
        """Check if the path exists."""

    @abstractmethod
    def mkdir(self, parents: bool = False, exist_ok: bool = False) -> None:
        """Create a directory."""

    @abstractmethod
    def absolute(self) -> StoragePath:
        """Return an absolute version of this path"""

    def resolve(self) -> StoragePath:
        """Resolve the resolved path of the path."""
        raise NotImplementedError(f"resolve in {self.__class__} is not implemented yet.")

    @abstractmethod
    def glob(self, pattern: str) -> Generator[StoragePath, StoragePath, StoragePath]:
        """Iterate over this subtree and yield all existing files"""

    @abstractmethod
    def __truediv__(self, key: str) -> StoragePath:
        """Return StoragePath with appending the key to the exising path"""



"""RemoveHideHeaderFooterStatements migration"""

from typing import ClassVar

from cmem_cmemc.migrations.abc import MigrationRecipe, components


class RemoveHideHeaderFooterStatements(MigrationRecipe):
    """25.1 Remove Non-Operational shui:valueQueryHideHeader|Footer Triple"""

    id = "hide-header-footer-25.1"
    description = "Remove triples using deprecated shui:valueQueryHideHeader|Footer terms"
    component: components = "explore"
    first_version = "25.1"
    tags: ClassVar[list[str]] = ["shapes", "user"]
    check_query = """{{DEFAULT_PREFIXES}}
SELECT ?shape
WHERE {
  GRAPH ?shapeCatalog {
    ?shapeCatalog a shui:ShapeCatalog .
    ?shape ?deprecatedProperty ?value .
  }
  VALUES ?deprecatedProperty {
    shui:valueQueryHideHeader shui:valueQueryHideFooter
  }
} LIMIT 1
    """
    delete_query = """{{DEFAULT_PREFIXES}}
DELETE {
  GRAPH ?shapeCatalog { ?shape ?deprecatedProperty ?value . }
}
WHERE {
  GRAPH ?shapeCatalog {
    ?shapeCatalog a shui:ShapeCatalog .
    ?shape ?deprecatedProperty ?value .
  }
  VALUES ?deprecatedProperty {
    shui:valueQueryHideHeader shui:valueQueryHideFooter
  }
}
    """

    def is_applicable(self) -> bool:
        """Test if the recipe can be applied."""
        existing_triples = self._select(self.check_query)
        return len(existing_triples) == 1

    def apply(self) -> None:
        """Apply the recipe to the current version."""
        self._update(self.delete_query)



"""Custom Click smart_path ParamType"""

import os
from typing import IO, Any

import click
import smart_open
from click.core import Context, Parameter
from smart_open import compression

from cmem_cmemc.context import ApplicationContext


class ClickSmartPath(click.Path):
    """Custom Click smart_path ParamType"""

    name = "click-smart-path"

    def __init__(  # noqa: PLR0913
        self,
        exists: bool = False,
        file_okay: bool = True,
        dir_okay: bool = True,
        writable: bool = False,
        readable: bool = True,
        resolve_path: bool = False,
        allow_dash: bool = False,
        remote_okay: bool = False,
    ):
        super().__init__(
            exists=exists,
            file_okay=file_okay,
            dir_okay=dir_okay,
            writable=writable,
            readable=readable,
            resolve_path=resolve_path,
            allow_dash=allow_dash,
        )
        self.remote_okay = remote_okay

    def convert(
        self,
        value: str | os.PathLike[str],
        param: Parameter | None,
        ctx: Context | None,
    ) -> str | bytes | os.PathLike[str]:
        """Convert the given value"""
        try:
            parsed_path = smart_open.parse_uri(value)
        except NotImplementedError as exe:
            self.fail(f"{exe}", param, ctx)
        if parsed_path.scheme == "file":
            return super().convert(parsed_path.uri_path, param, ctx)
        if not self.remote_okay:
            self.fail("Remote path not supported", param, ctx)

        return value

    @staticmethod
    def open(
        file_path: str, mode: str = "rb", transport_params: dict[str, Any] | None = None
    ) -> IO:
        """Open the file and return the file handle."""
        if file_path == "-":
            return click.open_file(file_path, mode=mode)
        if transport_params is None:
            transport_params = {}
        transport_params["timeout"] = os.getenv(
            "CMEM_TRANSPORT_TIMEOUT", ApplicationContext.DEFAULT_EXTERNAL_HTTP_TIMEOUT
        )
        return smart_open.open(  # type: ignore[no-any-return]
            file_path,
            mode,
            transport_params=transport_params,
            compression=compression.NO_COMPRESSION,
        )



"""admin commands for cmem command line interface."""

from datetime import datetime, timezone

import click
import jwt
from click import ClickException
from cmem.cmempy.api import get_access_token, get_token
from cmem.cmempy.config import get_cmem_base_uri
from cmem.cmempy.health import get_complete_status_info
from dateutil.relativedelta import relativedelta

from cmem_cmemc import completion
from cmem_cmemc.command import CmemcCommand
from cmem_cmemc.command_group import CmemcGroup
from cmem_cmemc.commands.acl import acl
from cmem_cmemc.commands.client import client
from cmem_cmemc.commands.metrics import metrics
from cmem_cmemc.commands.migration import migration
from cmem_cmemc.commands.store import store
from cmem_cmemc.commands.user import user
from cmem_cmemc.commands.workspace import workspace
from cmem_cmemc.context import ApplicationContext
from cmem_cmemc.utils import struct_to_table

WARNING_MIGRATION = (
    "Your workspace configuration version does not match your Explore version. "
    "Please consider migrating your workspace configuration (admin store migrate command)."
)
WARNING_SHAPES = (
    "Your ShapeCatalog version does not match your Explore version. "
    "Please consider updating your bootstrap data (admin store boostrap command)."
)


def _check_cmem_license(app: ApplicationContext, data: dict, exit_1: str) -> None:
    """Check grace period of CMEM license."""
    if "license" not in data["explore"]["info"]:
        # DP < 24.1 has no cmem license information here
        return
    license_ = data["explore"]["info"]["license"]
    in_grace_period: bool = license_.get("inGracePeriod", False)
    if in_grace_period:
        cmem_license_end = license_["validDate"]
        output = f"Your Corporate Memory license expired on {cmem_license_end}."
        if exit_1 in ("error", "always"):
            raise ClickException(output)
        app.echo_error(output)


def _check_graphdb_license(app: ApplicationContext, data: dict, months: int, exit_1: str) -> None:
    """Check grace period of graphdb license."""
    if "licenseExpiration" not in data["explore"]["info"]["store"]:
        # DP < 24.1 has no graph license information here
        return
    expiration_date_str = data["explore"]["info"]["store"]["licenseExpiration"]
    expiration_date = datetime.strptime(expiration_date_str, "%Y-%m-%d").astimezone(tz=timezone.utc)
    grace_starts = expiration_date - relativedelta(months=months)
    if grace_starts < datetime.now(tz=timezone.utc):
        graphdb_license_end = data["explore"]["info"]["store"]["licenseExpiration"]
        output = f"Your GraphDB license expires on {graphdb_license_end}."
        if exit_1 == "always":
            raise ClickException(output)
        app.echo_warning(output)


@click.command(cls=CmemcCommand, name="status")
@click.option(
    "--key",
    "key",
    shell_complete=completion.status_keys,
    help="Get only specific key(s) from the status / info output. There are "
    "two special keys available: 'all' will list all available keys in "
    "the table, 'overall.healthy' with result in  UP in case all "
    "health flags are UP as well (otherwise DOWN).",
)
@click.option(
    "--exit-1",
    type=click.Choice(["never", "error", "always"]),
    default="never",
    show_default=True,
    help="Specify, when this command returns with exit code 1. Available options are "
    "'never' (exit 0 on errors and warnings), "
    "'error' (exit 1 on errors, exit 0 on warnings), "
    "'always' (exit 1 on errors and warnings).",
)
@click.option(
    "--enforce-table",
    is_flag=True,
    help="A single value with --key will be returned as plain text instead "
    "of a table with one row and the header. This default behaviour "
    "allows for more easy integration with scripts. This flag enforces "
    "the use of tabular output, even for single row tables.",
)
@click.option(
    "--raw", is_flag=True, help="Outputs combined raw JSON output of the health/info endpoints."
)
@click.pass_obj
def status_command(  # noqa: C901, PLR0912
    app: ApplicationContext, key: str, exit_1: str, enforce_table: bool, raw: bool
) -> None:
    """Output health and version information.

    This command outputs version and health information of the
    selected deployment. If the version information cannot be retrieved,
    UNKNOWN is shown.

    Additionally, this command informs you in one of these cases:
    (1) A warning, if the target version of your cmemc client is newer than the version
    of your backend.
    (2) A warning, if the ShapeCatalog has a different version than your Explore component.
    (3) An error, if your Corporate Memory license is expired (grace period).
    (4) A warning, if your Graph DB license will expire in less than a month.

    To get status information of all configured
    deployments use this command in combination with parallel.

    Example: cmemc config list | parallel --ctag cmemc -c {} admin status
    """
    _ = get_complete_status_info()
    if "error" in _["build"]:
        app.echo_debug(_["build"]["error"])
    if "error" in _["explore"]:
        app.echo_debug(_["explore"]["error"])

    if exit_1 in ("always", "error") and (_["overall"]["healthy"] != "UP"):
        raise ClickException(
            f"One or more major status flags are DOWN or UNKNOWN: {_!r}",
        )
    if raw:
        app.echo_info_json(_)
        return
    if key:
        table = [line for line in struct_to_table(_) if line[0].startswith(key) or key == "all"]
        if len(table) == 1 and not enforce_table:
            app.echo_info(table[0][1])
            return
        if len(table) == 0:
            raise ClickException(f"No values for key(s): {key}")
        app.echo_info_table(table, headers=["Key", "Value"], sort_column=0)
        return
    app.check_versions()
    _workspace_config = _["explore"]["info"].get("workspaceConfiguration", {})
    if _workspace_config.get("workspacesToMigrate"):
        if exit_1 == "always":
            raise ClickException(WARNING_MIGRATION)
        app.echo_warning(WARNING_MIGRATION)

    if _["shapes"]["version"] not in (_["explore"]["version"], "UNKNOWN"):
        if exit_1 == "always":
            raise ClickException(WARNING_SHAPES)
        app.echo_warning(WARNING_SHAPES)

    _check_cmem_license(app=app, data=_, exit_1=exit_1)
    _check_graphdb_license(app=app, data=_, months=1, exit_1=exit_1)
    if _["store"]["type"] != "GRAPHDB":
        store_version = _["store"]["type"] + "/" + _["store"]["version"]
    else:
        store_version = _["store"]["version"]
    table = [
        ("EXPLORE", _["explore"]["version"], _["explore"]["healthy"]),
        ("BUILD", _["build"]["version"], _["build"]["healthy"]),
        ("SHAPES", _["shapes"]["version"], _["shapes"]["healthy"]),
        ("STORE", store_version, _["store"]["healthy"]),
    ]
    app.echo_info_table(
        table,
        headers=["Component", "Version", "Status"],
        sort_column=0,
        caption=f"Status of {get_cmem_base_uri()}",
    )


@click.command(cls=CmemcCommand, name="token")
@click.option(
    "--raw",
    is_flag=True,
    help="Outputs raw JSON. Note that this option will always try to fetch "
    "a new JSON token response. In case you are working with "
    "OAUTH_GRANT_TYPE=prefetched_token, this may lead to an error.",
)
@click.option(
    "--decode",
    is_flag=True,
    help="Decode the access token and outputs the raw JSON. Note that the "
    "access token is only decoded and esp. not validated.",
)
@click.pass_obj
def token_command(app: ApplicationContext, raw: bool, decode: bool) -> None:
    """Fetch and output an access token.

    This command can be used to check for correct authentication as well as
    to use the token with wget / curl or similar standard tools:

    Example: curl -H "Authorization: Bearer $(cmemc -c my admin token)"
    $(cmemc -c my config get DP_API_ENDPOINT)/api/custom/slug

    Please be aware that this command can reveal secrets which you might
    not want to be present in log files or on the screen.
    """
    # Note:
    # - get_access_token returns the token string which is maybe from conf
    # - get_token fetches a new token incl. envelope from keycloak

    if decode:
        token = get_access_token()
        decoded_token = jwt.decode(token, options={"verify_signature": False})
        if raw:
            app.echo_info_json(decoded_token)
            return
        table = struct_to_table(decoded_token)
        app.echo_info_table(table, headers=["Key", "Value"], sort_column=0)
        return
    if raw:
        app.echo_info_json(get_token())
        return
    token = get_access_token()
    app.echo_info(token)


@click.group(cls=CmemcGroup)
def admin() -> CmemcGroup:  # type: ignore[empty-body]
    """Import bootstrap data, backup/restore workspace or get status.

    This command group consists of commands for setting up and
    configuring eccenca Corporate Memory.
    """


admin.add_command(status_command)
admin.add_command(token_command)
admin.add_command(metrics)
admin.add_command(workspace)
admin.add_command(store)
admin.add_command(user)
admin.add_command(acl)
admin.add_command(client)
admin.add_command(migration)



"""access control"""

import click
import requests.exceptions
from click import Option
from cmem.cmempy.dp.authorization.conditions import (
    create_access_condition,
    delete_access_condition,
    fetch_all_acls,
    get_access_condition_by_iri,
    review_graph_rights,
    update_access_condition,
)
from cmem.cmempy.keycloak.user import get_user_by_username, user_groups

from cmem_cmemc import completion
from cmem_cmemc.command import CmemcCommand
from cmem_cmemc.command_group import CmemcGroup
from cmem_cmemc.constants import NS_ACL, NS_ACTION, NS_GROUP, NS_USER
from cmem_cmemc.context import ApplicationContext
from cmem_cmemc.utils import (
    convert_iri_to_qname,
    convert_qname_to_iri,
    get_query_text,
    struct_to_table,
)

# option descriptions
HELP_TEXTS = {
    "name": "A optional name.",
    "id": "An optional ID (will be an UUID otherwise).",
    "description": "An optional description.",
    "user": "A specific user account required by the access condition.",
    "group": "A membership in a user group required by the access condition.",
    "read_graph": "Grants read access to a graph.",
    "write_graph": "Grants write access to a graph (includes read access).",
    "action": "Grants usage permissions to an action / functionality.",
    "query": "Dynamic access condition query (file or the query catalog IRI).",
}

WARNING_UNKNOWN_USER = "Unknown User or no access to get user info."
WARNING_NO_GROUP_ACCESS = "You do not have the permission to retrieve user groups"
WARNING_USE_GROUP = "Use the --group option to assign groups manually (what-if-scenario)."

PUBLIC_USER_URI = "https://vocab.eccenca.com/auth/AnonymousUser"
PUBLIC_GROUP_URI = "https://vocab.eccenca.com/auth/PublicGroup"

KNOWN_ACCESS_CONDITION_URLS = [PUBLIC_USER_URI, PUBLIC_GROUP_URI]


def _list_to_acl_url(ctx: ApplicationContext, param: Option, value: list) -> list:
    """Option callback which returns a URI for a list of strings.

    or list of URIs, if a tuple comes from click
    or not, if it is already a known URI .... or None
    """
    return [_value_to_acl_url(ctx, param, _) for _ in value]


def _value_to_acl_url(
    ctx: ApplicationContext,  # noqa: ARG001
    param: Option,
    value: str | None,
) -> str | None:
    """Option callback which returns a URI for a string.

    or not, if it is already a known URI .... or None
    """
    if value == "" or value is None:
        return value
    if value.startswith(("http://", "https://")):
        return value
    match param.name:
        case "groups":
            return f"{NS_GROUP}{value}"
        case "user":
            return f"{NS_USER}{value}"
    return f"{NS_ACL}{value}"


def generate_acl_name(user: str | None, groups: list[str], query: str | None) -> str:
    """Create an access condition name based on user and group assignments."""
    if query is not None:
        return "Query based Dynamic Access Condition"
    if len(groups) > 0:
        group_term = "groups" if len(groups) > 1 else "group"
        groups_labels = ", ".join(
            [convert_iri_to_qname(iri=_, default_ns=NS_GROUP)[1:] for _ in groups]
        )
        if user:
            return (
                f"Condition for user {convert_iri_to_qname(iri=user, default_ns=NS_USER)[1:]} "
                f"and {group_term} {groups_labels}"
            )
        return f"Condition for {group_term} {groups_labels}"
    if user:
        return f"Condition for user: {convert_iri_to_qname(iri=user, default_ns=NS_USER)[1:]}"
    return "Condition for ALL users"


@click.command(cls=CmemcCommand, name="list")
@click.option("--raw", is_flag=True, help="Outputs raw JSON.")
@click.option(
    "--id-only",
    is_flag=True,
    help="Lists only URIs. This is useful for piping the IDs into other commands.",
)
@click.pass_obj
def list_command(app: ApplicationContext, raw: bool, id_only: bool) -> None:
    """List access conditions.

    This command retrieves and lists all access conditions, which are manageable
    by the current account.
    """
    acls = fetch_all_acls()
    if raw:
        app.echo_info_json(acls)
        return
    if id_only:
        for graph in acls:
            app.echo_info(convert_iri_to_qname(iri=graph.get("iri"), default_ns=NS_ACL))
        return
    table = [
        (convert_iri_to_qname(iri=_.get("iri"), default_ns=NS_ACL), _.get("name", "-"))
        for _ in acls
    ]
    app.echo_info_table(
        table,
        headers=["URI", "Name"],
        sort_column=0,
        empty_table_message="No access conditions found. "
        "Use the `admin acl create` command to create a new access condition.",
    )


@click.command(cls=CmemcCommand, name="inspect")
@click.argument("access_condition_id", type=click.STRING, shell_complete=completion.acl_ids)
@click.option("--raw", is_flag=True, help="Outputs raw JSON.")
@click.pass_obj
def inspect_command(app: ApplicationContext, access_condition_id: str, raw: bool) -> None:
    """Inspect an access condition.

    Note: access conditions can be listed by using the `acl list` command.
    """
    iri = convert_qname_to_iri(qname=access_condition_id, default_ns=NS_ACL)
    access_condition = get_access_condition_by_iri(iri).json()

    if raw:
        app.echo_info_json(access_condition)
        return

    table = struct_to_table(access_condition)
    app.echo_info_table(table, headers=["Key", "Value"], sort_column=0)


@click.command(cls=CmemcCommand, name="create")
@click.option(
    "--user",
    type=click.STRING,
    shell_complete=completion.acl_users,
    help=HELP_TEXTS["user"],
    callback=_value_to_acl_url,
)
@click.option(
    "--group",
    "groups",
    type=click.STRING,
    multiple=True,
    shell_complete=completion.acl_groups,
    help=HELP_TEXTS["group"],
    callback=_list_to_acl_url,
)
@click.option(
    "--read-graph",
    "read_graphs",
    type=click.STRING,
    multiple=True,
    shell_complete=completion.graph_uris_with_all_graph_uri,
    help=HELP_TEXTS["read_graph"],
)
@click.option(
    "--write-graph",
    "write_graphs",
    type=click.STRING,
    multiple=True,
    shell_complete=completion.graph_uris_with_all_graph_uri,
    help=HELP_TEXTS["write_graph"],
)
@click.option(
    "--action",
    "actions",
    type=click.STRING,
    multiple=True,
    shell_complete=completion.acl_actions,
    help=HELP_TEXTS["action"],
)
@click.option(
    "--query",
    "query",
    type=click.STRING,
    shell_complete=completion.remote_queries_and_sparql_files,
    help=HELP_TEXTS["query"],
)
@click.option(
    "--id",
    "id_",
    type=click.STRING,
    help=HELP_TEXTS["id"],
)
@click.option(
    "--name",
    "name",
    type=click.STRING,
    help=HELP_TEXTS["name"],
)
@click.option(
    "--description",
    "description",
    type=click.STRING,
    help=HELP_TEXTS["description"],
)
@click.pass_obj
# pylint: disable-msg=too-many-arguments
def create_command(  # noqa: PLR0913
    app: ApplicationContext,
    name: str,
    id_: str,
    description: str,
    user: str,
    groups: list[str],
    read_graphs: tuple[str],
    write_graphs: tuple[str],
    actions: tuple[str],
    query: str,
) -> None:
    """Create an access condition.

    With this command, new access conditions can be created.

    An access condition captures information about WHO gets access to WHAT.
    In order to specify WHO gets access, use the `--user` and / or `--group` options.
    In order to specify WHAT an account get access to, use the `--read-graph`,
    `--write-graph` and `--action` options.`

    In addition to that, you can specify a name, a description and an ID (all optional).

    A special case are dynamic access conditions, based on a SPARQL query: Here you
    have to provide a query with the projection variables `user`, `group` `readGraph`
    and `writeGraph` to create multiple grants at once. You can either provide a query file
    or a query URL from the query catalog.

    Note: Queries for dynamic access conditions are copied into the ACL, so changing the
    query in the query catalog does not change it in the access condition.

    Example: cmemc admin acl create --group local-users --write-graph https://example.org/
    """
    if not read_graphs and not write_graphs and not actions and not query:
        raise click.UsageError(
            "Missing access / usage grant. Use at least one of the following options: "
            "--read-graph, --write-graph, --action or --query."
        )
    query_str = None
    if query:
        query_str = get_query_text(query, {"user", "group", "readGraph", "writeGraph"})

    if not user and not groups and not query:
        app.echo_warning("Access conditions without a user or group assignment affect ALL users.")

    if not name:
        name = generate_acl_name(user=user, groups=groups, query=query)

    if not description:
        description = "This access condition was created with cmemc."

    app.echo_info(
        f"Creating access condition '{name}' ... ",
        nl=False,
    )

    create_access_condition(
        name=name,
        static_id=id_,
        description=description,
        user=user,
        groups=groups,
        read_graphs=list(read_graphs),
        write_graphs=list(write_graphs),
        actions=[convert_qname_to_iri(qname=_, default_ns=NS_ACTION) for _ in actions],
        query=query_str,
    )
    app.echo_success("done")


@click.command(cls=CmemcCommand, name="update")
@click.argument(
    "access_condition_id",
    nargs=1,
    required=True,
    type=click.STRING,
    shell_complete=completion.acl_ids,
)
@click.option(
    "--name",
    "name",
    type=click.STRING,
    help=HELP_TEXTS["name"],
)
@click.option(
    "--description",
    "description",
    type=click.STRING,
    help=HELP_TEXTS["description"],
)
@click.option(
    "--user",
    type=click.STRING,
    shell_complete=completion.acl_users,
    help=HELP_TEXTS["user"],
    callback=_value_to_acl_url,
)
@click.option(
    "--group",
    "groups",
    type=click.STRING,
    multiple=True,
    shell_complete=completion.acl_groups,
    help=HELP_TEXTS["group"],
    callback=_list_to_acl_url,
)
@click.option(
    "--read-graph",
    "read_graphs",
    type=click.STRING,
    multiple=True,
    shell_complete=completion.graph_uris_with_all_graph_uri,
    help=HELP_TEXTS["read_graph"],
)
@click.option(
    "--write-graph",
    "write_graphs",
    type=click.STRING,
    multiple=True,
    shell_complete=completion.graph_uris_with_all_graph_uri,
    help=HELP_TEXTS["write_graph"],
)
@click.option(
    "--action",
    "actions",
    type=click.STRING,
    multiple=True,
    shell_complete=completion.acl_actions,
    help=HELP_TEXTS["action"],
)
@click.option(
    "--query",
    "query",
    type=click.STRING,
    shell_complete=completion.remote_queries_and_sparql_files,
    help=HELP_TEXTS["query"],
)
@click.pass_obj
# pylint: disable-msg=too-many-arguments
def update_command(  # noqa: PLR0913
    app: ApplicationContext,
    access_condition_id: str,
    name: str,
    description: str,
    user: str,
    groups: list[str],
    read_graphs: tuple[str],
    write_graphs: tuple[str],
    actions: tuple[str],
    query: str,
) -> None:
    """Update an access condition.

    Given an access condition URL, you can change specific options
    to new values.
    """
    iri = convert_qname_to_iri(qname=access_condition_id, default_ns=NS_ACL)
    payload = get_access_condition_by_iri(iri=iri).json()
    app.echo_info(
        f"Updating access condition {payload['name']} ... ",
        nl=False,
    )
    query_str = None
    if query:
        query_str = get_query_text(query, {"user", "group", "readGraph", "writeGraph"})

    update_access_condition(
        iri=iri,
        name=name,
        description=description,
        user=user,
        groups=groups,
        read_graphs=read_graphs,
        write_graphs=write_graphs,
        actions=[convert_qname_to_iri(qname=_, default_ns=NS_ACTION) for _ in actions],
        query=query_str,
    )
    app.echo_success("done")


@click.command(cls=CmemcCommand, name="delete")
@click.option(
    "-a",
    "--all",
    "all_",
    is_flag=True,
    help="Delete all access conditions. " "This is a dangerous option, so use it with care.",
)
@click.argument(
    "access_condition_ids",
    nargs=-1,
    type=click.STRING,
    shell_complete=completion.acl_ids,
)
@click.pass_obj
def delete_command(app: ApplicationContext, all_: bool, access_condition_ids: list[str]) -> None:
    """Delete access conditions.

    This command deletes existing access conditions from the account.

    Note: Access conditions can be listed by using the `cmemc admin acs list` command.
    """
    if access_condition_ids == () and not all_:
        raise click.UsageError(
            "Either specify at least one access condition ID,"
            " or use the --all option to delete all access conditions."
        )
    if all_:
        access_condition_ids = [_["iri"] for _ in fetch_all_acls()]

    count = len(access_condition_ids)
    for index, _ in enumerate(access_condition_ids, 1):
        app.echo_info(f"Delete access condition {index}/{count}: {_} ... ", nl=False)
        delete_access_condition(iri=convert_qname_to_iri(qname=_, default_ns=NS_ACL))
        app.echo_success("done")


@click.command(name="review")
@click.option("--raw", is_flag=True, help="Outputs raw JSON.")
@click.argument("user", type=click.STRING, shell_complete=completion.acl_users)
@click.option(
    "--group",
    "groups",
    type=click.STRING,
    multiple=True,
    shell_complete=completion.acl_groups,
    callback=_list_to_acl_url,
    help="Add groups to the review request (what-if-scenario).",
)
@click.pass_obj
def review_command(app: ApplicationContext, raw: bool, user: str, groups: list[str] | None) -> None:
    """Review grants for a given account.

    This command has two working modes: (1) You can review the access conditions
    of an actual account,
    (2) You can review the access conditions of an imaginary account with a set of
    freely added groups (what-if-scenario).

    The output of the command is a list of grants the account has based on your input
    and all access conditions loaded in the store. In addition to that, some metadata
    of the account is shown.
    """
    if not groups:
        app.echo_debug("Trying to fetch groups from keycloak.")
        keycloak_user = get_user_by_username(username=user)
        if not keycloak_user:
            if user != PUBLIC_USER_URI:
                app.echo_warning(WARNING_UNKNOWN_USER)
                app.echo_warning(WARNING_USE_GROUP)
        else:
            try:
                keycloak_user_groups = user_groups(user_id=keycloak_user[0]["id"])
                groups = [f"{NS_GROUP}{_['name']}" for _ in keycloak_user_groups]
            except (requests.exceptions.HTTPError, IndexError):
                app.echo_warning(WARNING_NO_GROUP_ACCESS)
                app.echo_warning(WARNING_USE_GROUP)
    app.echo_debug(f"Got groups: {groups}")
    account_iri = f"{NS_USER}{user}" if user != PUBLIC_USER_URI else PUBLIC_USER_URI
    review_info: dict = review_graph_rights(account_iri=account_iri, group_iris=groups).json()
    review_info["groupIri"] = groups
    if raw:
        app.echo_info_json(review_info)
        return
    table = struct_to_table(review_info)
    app.echo_info_table(table, headers=["Key", "Value"], sort_column=0)


@click.group(cls=CmemcGroup)
def acl() -> CmemcGroup:  # type: ignore[empty-body]
    """List, create, delete and modify and review access conditions.

    With this command group, you can manage and inspect access conditions
    in eccenca Corporate Memory. Access conditions are identified by a URL.
    They grant access to knowledge graphs or actions to user or groups.
    """


acl.add_command(list_command)
acl.add_command(inspect_command)
acl.add_command(create_command)
acl.add_command(update_command)
acl.add_command(delete_command)
acl.add_command(review_command)



"""Access Conditions migration recipes from before 24.3"""

from typing import ClassVar

from cmem.cmempy.dp.authorization.refresh import get as refresh_acls

from cmem_cmemc.migrations.abc import MigrationRecipe, components


class MoveAccessConditionsToNewGraph(MigrationRecipe):
    """24.4 Access Conditions Migration"""

    id = "acl-graph-24.3"
    description = "Move access conditions and used queries to new ACL graph"
    component: components = "explore"
    first_version = "24.3"
    tags: ClassVar[list[str]] = ["system", "acl"]
    check_query = """{{DEFAULT_PREFIXES}}
SELECT ?subject
FROM <urn:elds-backend-access-conditions-graph>
WHERE {
  ?subject a ?class .
  FILTER NOT EXISTS { ?subject shui:isSystemResource true }
  FILTER (?class IN (auth:AccessCondition, shui:SparqlQuery) ) .
}
    """
    move_query = """{{DEFAULT_PREFIXES}}
DELETE { GRAPH <urn:elds-backend-access-conditions-graph> { ?s ?p ?o . } }
INSERT { GRAPH <https://ns.eccenca.com/data/ac/> { ?s ?p ?o . } }
WHERE {
  GRAPH <urn:elds-backend-access-conditions-graph> {
    ?s ?p ?o .
    ?s a ?class .
    FILTER (?class IN (auth:AccessCondition, shui:SparqlQuery) ) .
    FILTER NOT EXISTS { ?s shui:isSystemResource true }
  }
}
    """

    def is_applicable(self) -> bool:
        """Test if the recipe can be applied."""
        old_acls = self._select(self.check_query)
        return len(old_acls) > 0

    def apply(self) -> None:
        """Apply the recipe to the current version."""
        self._update(self.move_query)
        refresh_acls()


class RenameAuthVocabularyResources(MigrationRecipe):
    """24.4 Access Conditions Migration"""

    id = "acl-vocab-24.3"
    description = "Migrate auth vocabulary terms (actions and other grants)"
    component: components = "explore"
    first_version = "24.3"
    tags: ClassVar[list[str]] = ["system", "acl"]
    check_query = """{{DEFAULT_PREFIXES}}
SELECT DISTINCT ?acl_id
WHERE {
  GRAPH ?acl_graph {
	?acl_id a auth:AccessCondition .
    FILTER NOT EXISTS { ?acl_id shui:isSystemResource true }
    ?acl_id ?property ?old_term .
    FILTER (?old_term in (
        <urn:elds-backend-all-actions>,
        <urn:elds-backend-actions-auth-access-control>,
        <urn:eccenca:di>,
        <urn:eccenca:ThesaurusUserInterface>,
        <urn:eccenca:AccessInternalGraphs>,
        <urn:eccenca:QueryUserInterface>,
        <urn:eccenca:VocabularyUserInterface>,
        <urn:eccenca:ExploreUserInterface>,
        <https://vocab.eccenca.com/auth/Action/Viz/Manage>,
        <https://vocab.eccenca.com/auth/Action/Viz/Read>,
        <urn:elds-backend-anonymous-user>,
        <urn:elds-backend-public-group>,
        <urn:elds-backend-all-graphs>
      ))
  }
  FILTER (?acl_graph IN (<urn:elds-backend-access-conditions-graph>, <https://ns.eccenca.com/data/ac/>))
}
    """
    move_query = """{{DEFAULT_PREFIXES}}
DELETE { GRAPH ?acl_graph { ?s ?p <{{OLD_IRI}}> . } }
INSERT { GRAPH ?acl_graph { ?s ?p <{{NEW_IRI}}> . } }
WHERE {
  GRAPH ?acl_graph {
    ?s ?p ?o .
    FILTER NOT EXISTS { ?s shui:isSystemResource true }
    FILTER (?o = <{{OLD_IRI}}> )
  }
  FILTER (?acl_graph IN (<urn:elds-backend-access-conditions-graph>, <https://ns.eccenca.com/data/ac/>))
}
    """
    terms: ClassVar[dict[str, str]] = {
        "urn:elds-backend-all-actions": "https://vocab.eccenca.com/auth/Action/AllActions",
        "urn:elds-backend-actions-auth-access-control": "https://vocab.eccenca.com/auth/Action/ChangeAccessConditions",
        "urn:eccenca:di": "https://vocab.eccenca.com/auth/Action/ChangeAccessConditions",
        "urn:eccenca:ThesaurusUserInterface": "https://vocab.eccenca.com/auth/Action/Explore-ThesaurusCatalog",
        "urn:eccenca:AccessInternalGraphs": "https://vocab.eccenca.com/auth/Action/Explore-ListSystemGraphs",
        "urn:eccenca:QueryUserInterface": "https://vocab.eccenca.com/auth/Action/Explore-QueryCatalog",
        "urn:eccenca:VocabularyUserInterface": "https://vocab.eccenca.com/auth/Action/Explore-VocabularyCatalog",
        "urn:eccenca:ExploreUserInterface": "https://vocab.eccenca.com/auth/Action/Explore-KnowledgeGraphs",
        "https://vocab.eccenca.com/auth/Action/Viz/Manage": "https://vocab.eccenca.com/auth/Action/Explore-BKE-Manage",
        "https://vocab.eccenca.com/auth/Action/Viz/Read": "https://vocab.eccenca.com/auth/Action/Explore-BKE-Read",
        "urn:elds-backend-anonymous-user": "https://vocab.eccenca.com/auth/AnonymousUser",
        "urn:elds-backend-public-group": "https://vocab.eccenca.com/auth/PublicGroup",
        "urn:elds-backend-all-graphs": "https://vocab.eccenca.com/auth/AllGraphs",
    }

    def is_applicable(self) -> bool:
        """Test if the recipe can be applied."""
        acls_with_vocabs_to_change = self._select(self.check_query)
        return len(acls_with_vocabs_to_change) > 0

    def apply(self) -> None:
        """Apply the recipe to the current version."""
        for old_iri, new_iri in self.terms.items():
            self._update(self.move_query, placeholder={"OLD_IRI": old_iri, "NEW_IRI": new_iri})
        refresh_acls()



"""Workspace Configuration migration recipe"""

from typing import ClassVar

from cmem.cmempy.dp.workspace import migrate_workspaces
from cmem.cmempy.health import get_complete_status_info

from cmem_cmemc.migrations.abc import MigrationRecipe, components


class MigrateWorkspaceConfiguration(MigrationRecipe):
    """Update Workspace Configuration to current version"""

    id = "workspace-configurations"
    description = "Migrate explore workspace configurations to the current version"
    component: components = "explore"
    first_version = "24.2"
    tags: ClassVar[list[str]] = ["system", "config"]

    def is_applicable(self) -> bool:
        """Test if the recipe can be applied."""
        status_info = get_complete_status_info()
        config = status_info["explore"]["info"].get("workspaceConfiguration", {})
        return bool(config.get("workspacesToMigrate"))

    def apply(self) -> None:
        """Apply the recipe to the current version."""
        migrate_workspaces()



"""cmemc Click Command Group"""

from click_didyoumean import DYMGroup
from click_help_colors import HelpColorsGroup


class CmemcGroup(HelpColorsGroup, DYMGroup):
    """Wrapper click.Group class to have a single extension point.

    Currently, wrapped click extensions and additional group features:#
    - click-help-colors: https://github.com/click-contrib/click-help-colors
    """

    color_for_command_groups = "white"
    color_for_writing_commands = "red"
    color_for_headers = "yellow"
    color_for_options = "green"

    def __init__(self, *args, **kwargs):  # noqa: ANN002, ANN003
        """Init a cmemc group command group."""
        # set default colors
        kwargs.setdefault("help_headers_color", self.color_for_headers)
        kwargs.setdefault("help_options_color", self.color_for_options)
        kwargs.setdefault(
            "help_options_custom_colors",
            {
                "bootstrap": self.color_for_writing_commands,
                "showcase": self.color_for_writing_commands,
                "delete": self.color_for_writing_commands,
                "password": self.color_for_writing_commands,
                "secret": self.color_for_writing_commands,
                "upload": self.color_for_writing_commands,
                "import": self.color_for_writing_commands,
                "create": self.color_for_writing_commands,
                "enable": self.color_for_writing_commands,
                "disable": self.color_for_writing_commands,
                "execute": self.color_for_writing_commands,
                "replay": self.color_for_writing_commands,
                "io": self.color_for_writing_commands,
                "install": self.color_for_writing_commands,
                "uninstall": self.color_for_writing_commands,
                "reload": self.color_for_writing_commands,
                "update": self.color_for_writing_commands,
                "eval": self.color_for_writing_commands,
                "cancel": self.color_for_writing_commands,
                "admin": self.color_for_command_groups,
                "user": self.color_for_command_groups,
                "store": self.color_for_command_groups,
                "metrics": self.color_for_command_groups,
                "config": self.color_for_command_groups,
                "dataset": self.color_for_command_groups,
                "graph": self.color_for_command_groups,
                "project": self.color_for_command_groups,
                "query": self.color_for_command_groups,
                "scheduler": self.color_for_command_groups,
                "vocabulary": self.color_for_command_groups,
                "workflow": self.color_for_command_groups,
                "workspace": self.color_for_command_groups,
                "python": self.color_for_command_groups,
                "cache": self.color_for_command_groups,
                "resource": self.color_for_command_groups,
                "acl": self.color_for_command_groups,
                "client": self.color_for_command_groups,
                "variable": self.color_for_command_groups,
                "validation": self.color_for_command_groups,
                "migrate": self.color_for_writing_commands,
                "migrations": self.color_for_command_groups,
                "imports": self.color_for_command_groups,
            },
        )
        super().__init__(*args, **kwargs)



"""Generate a help text and command structure graph."""

import contextlib

import click.core


def print_manual_graph(ctx: click.core.Context, version: str) -> None:
    """Output the complete manual graph.

    Returns: None
    """
    comment = "This dataset represents the complete inline documentation " "of cmemc as a graph"
    ctx.obj.echo_info(
        f"""
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>.
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#>.
@prefix void: <http://rdfs.org/ns/void#>.
@prefix xsd: <http://www.w3.org/2001/XMLSchema#>.
@prefix owl: <http://www.w3.org/2002/07/owl#>.
@prefix skos: <http://www.w3.org/2004/02/skos/core#>.
@prefix cli: <https://vocabs.eccenca.com/cli/>.
@prefix s: <http://schema.org/>.
@prefix : <https://cmemc.eccenca.dev/>.
@prefix i: <https://eccenca.com/go/cmemc>.

: a void:Dataset ;
    owl:versionInfo "{version}" ;
    rdfs:label "cmemc: Documentation Graph" ;
    rdfs:comment "{comment}" .

i: a s:SoftwareApplication ;
    rdfs:label "cmemc" ;
    s:softwareVersion "{version}" .
"""
    )
    print_group_manual_graph_recursive(ctx.command, ctx=ctx)


def print_group_manual_graph_recursive(
    command_group: click.Command | click.Group, ctx: click.core.Context, prefix: str = ""
) -> None:
    """Output documentation graph (recursive)."""
    commands = command_group.commands  # type: ignore[union-attr]
    for key in commands:
        if key == "manual":
            continue
        item = commands[key]
        iri = f":{prefix}{key}"
        ctx.obj.echo_info(f"{iri} skos:notation '{key}' .")
        if isinstance(item, click.Group):
            comment = item.get_short_help_str(limit=200)
            sub_group_iri = f":{prefix[:-1]}"
            if sub_group_iri == ":":
                sub_group_iri = "<https://eccenca.com/go/cmemc>"
            ctx.obj.echo_info(f"{iri} a cli:CommandGroup .")
            ctx.obj.echo_info(f"{iri} rdfs:label '{prefix}{key} Command Group' .")
            ctx.obj.echo_info(f"{iri} cli:subGroupOf {sub_group_iri} .")
            ctx.obj.echo_info(f"{iri} rdfs:comment '{comment}' .")
            print_group_manual_graph_recursive(item, ctx=ctx, prefix=f"{prefix}{key}-")
        elif isinstance(item, click.Command):
            comment = item.get_short_help_str(limit=200)
            group_iri = f":{prefix[:-1]}"
            ctx.obj.echo_info(f"{iri} a cli:Command .")
            ctx.obj.echo_info(f"{iri} rdfs:label '{prefix}{key} Command' .")
            ctx.obj.echo_info(f"{iri} cli:group {group_iri} .")
            ctx.obj.echo_info(f"{iri} rdfs:comment '{comment}' .")
            for parameter in item.params:
                print_parameter_manual_graph(parameter, ctx=ctx, prefix=f"{prefix}{key}-")
        else:
            pass


def print_parameter_manual_graph(
    item: click.Parameter | click.Option, ctx: click.core.Context, prefix: str = ""
) -> None:
    """Output documentation graph for parameter."""
    iri = f":{prefix}{item.name}"
    ctx.obj.echo_info(f"{iri} a cli:Parameter .")
    ctx.obj.echo_info(f"{iri} cli:command :{prefix[:-1]}.")
    if len(item.opts) == 1:
        ctx.obj.echo_info(f"{iri} rdfs:label '{item.opts[0]}'.")
    else:
        ctx.obj.echo_info(f'{iri} rdfs:label """{max(item.opts, key=len)}""".')
        ctx.obj.echo_info(f'{iri} cli:shortenedOption """{min(item.opts, key=len)}""".')
    if item.default not in (None, False, [None, None]):
        ctx.obj.echo_info(f"{iri} cli:defaultValue '{item.default}'.")
    with contextlib.suppress(AttributeError):
        ctx.obj.echo_info(f'{iri} rdfs:comment """{item.help}""".')  # type: ignore[union-attr]



"""The main command line interface."""

import os
import sys
import traceback
from importlib.resources import open_text
from os import environ as env

import click

from cmem_cmemc import completion
from cmem_cmemc.command_group import CmemcGroup
from cmem_cmemc.commands import (
    admin,
    config,
    dataset,
    graph,
    manual,
    project,
    query,
    vocabulary,
    workflow,
)
from cmem_cmemc.context import ApplicationContext
from cmem_cmemc.exceptions import CmemcError
from cmem_cmemc.utils import check_python_version, extract_error_message, get_version

CMEMC_VERSION = get_version()

# this will output a custom zsh completion function
if os.environ.get("_CMEMC_COMPLETE", "") == "zsh_source":
    with open_text("cmem_cmemc", "_cmemc.zsh") as zsh_output:
        click.secho(zsh_output.read())
    sys.exit(0)

version = sys.version_info
PYTHON_VERSION = f"{version.major}.{version.minor}.{version.micro}"
check_python_version(ctx=ApplicationContext)

# set the user-agent environment for the http request headers
os.environ["CMEM_USER_AGENT"] = f"cmemc/{CMEMC_VERSION} (Python {PYTHON_VERSION})"

# https://github.com/pallets/click/blob/master/examples/complex/complex/cli.py
CONTEXT_SETTINGS = {"auto_envvar_prefix": "CMEMC", "help_option_names": ["-h", "--help"]}


@click.group(name="cmemc", cls=CmemcGroup, context_settings=CONTEXT_SETTINGS)
@click.option(
    "-c",
    "--connection",
    type=click.STRING,
    shell_complete=completion.connections,
    help="Use a specific connection from the config file.",
)
@click.option(
    "--config-file",
    shell_complete=completion.ini_files,
    type=click.Path(readable=True, allow_dash=False, dir_okay=False),
    default=ApplicationContext.DEFAULT_CONFIG_FILE,
    show_default=f"Using {env['CMEMC_CONFIG_FILE']} "
    f"instead of {ApplicationContext.DEFAULT_CONFIG_FILE}"
    if "CMEMC_CONFIG_FILE" in env
    else ApplicationContext.DEFAULT_CONFIG_FILE,
    help="Use this config file instead of the default one.",
)
@click.option("-q", "--quiet", is_flag=True, help="Suppress any non-error info messages.")
@click.option(
    "-d", "--debug", is_flag=True, help="Output debug messages and stack traces after errors."
)
@click.option(
    "--external-http-timeout",
    default=ApplicationContext.DEFAULT_EXTERNAL_HTTP_TIMEOUT,
    type=int,
    show_default=True,
    help="Timeout in seconds for external HTTP requests.",
)
@click.version_option(
    version=CMEMC_VERSION,
    message="%(prog)s, version %(version)s, " f"running under python {PYTHON_VERSION}",
)
@click.pass_context
def cli(  # noqa: PLR0913
    ctx: click.core.Context,
    debug: bool,
    quiet: bool,
    config_file: str,
    connection: str,
    external_http_timeout: int,
) -> None:
    """Eccenca Corporate Memory Control (cmemc).

    cmemc is the eccenca Corporate Memory Command Line Interface (CLI).

    Available commands are grouped by affecting resource type (such as graph,
    project and query).
    Each command and group has a separate --help screen for detailed
    documentation.
    In order to see possible commands in a group, simply
    execute the group command without further parameter (e.g. cmemc project).

    If your terminal supports colors, these coloring rules are applied:
    Groups are colored in white; Commands which change data are colored in
    red; all other commands as well as options are colored in green.

    Please also have a look at the cmemc online documentation:

                        https://eccenca.com/go/cmemc

    cmemc is © 2025 eccenca GmbH, licensed under the Apache License 2.0.
    """
    _ = connection, debug, quiet, config_file, external_http_timeout
    if " ".join(sys.argv).find("config edit") != -1:
        app = ApplicationContext(config_file=config_file, debug=debug, quiet=quiet)
    else:
        app = ApplicationContext.from_params(params=ctx.params)
    ctx.obj = app


cli.add_command(admin.admin)
cli.add_command(config.config)
cli.add_command(dataset.dataset)
cli.add_command(graph.graph)
cli.add_command(project.project)
cli.add_command(query.query)
cli.add_command(vocabulary.vocabulary)
cli.add_command(workflow.workflow)
cli.add_command(manual.manual_command)


def main() -> None:
    """Start the command line interface."""
    try:
        cli()  # pylint: disable=no-value-for-parameter
    except CmemcError as error:
        app: ApplicationContext = error.app
        app.echo_debug(traceback.format_exc())
        app.echo_error(extract_error_message(error))
        sys.exit(1)



"""Declares all cli exceptions."""

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from cmem_cmemc.context import ApplicationContext


class CmemcError(ValueError):
    """Base exception for CMEM-CMEMC-related errors."""

    def __init__(self, app: "ApplicationContext", *args: str):
        super().__init__(*args)
        self.app = app


class InvalidConfigurationError(CmemcError):
    """The configuration given was not found or is broken."""


class ServerError(CmemcError):
    """The server reported an error with a process."""
